# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.stack.projection module.
"""

import pytest

import numpy as np
import apifish.stack as stack

from apifish.stack.projection import _one_hot_3d

from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal


# toy images
x_3d = np.array(
    [[[1, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[2, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[3, 0, 0, 0, 0],
      [0, 3, 0, 0, 3],
      [0, 3, 0, 0, 0],
      [0, 3, 3, 3, 0],
      [0, 0, 0, 0, 0]]],
    dtype=np.uint8)

x_3d_out_focus = np.array(
    [[[1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1]],

     [[1, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1]],

     [[2, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[3, 0, 0, 0, 0],
      [0, 3, 0, 0, 3],
      [0, 3, 0, 0, 0],
      [0, 3, 3, 3, 0],
      [0, 0, 0, 0, 0]],

     [[1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1]]],
    dtype=np.uint8)


# ### test 2-d projection ###

@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.float32, np.float64])
def test_maximum_projection(dtype):
    x = x_3d.astype(dtype)
    expected_y = np.array([[3, 0, 0, 0, 0],
                           [0, 3, 0, 0, 3],
                           [0, 3, 0, 0, 0],
                           [0, 3, 3, 3, 0],
                           [0, 0, 0, 0, 0]], dtype=dtype)
    y = stack.maximum_projection(x)
    assert_array_equal(y, expected_y)
    assert y.dtype == dtype


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.float32, np.float64])
def test_mean_projection(dtype):
    # 'return_float' == False
    x = x_3d.astype(dtype)
    expected_y = np.array([[2,   0,   0,   0,   0],
                           [0,   5/3, 0,   0,   1],
                           [0,   5/3, 0,   0,   0],
                           [0,   5/3, 5/3, 5/3, 0],
                           [0,   0,   0,   0,   0]], dtype=dtype)
    y = stack.mean_projection(x)
    assert_array_equal(y, expected_y)
    assert y.dtype == dtype

    # 'return_float' == True
    expected_y = np.array([[2, 0,   0,   0,   0],
                           [0, 5/3, 0,   0,   1],
                           [0, 5/3, 0,   0,   0],
                           [0, 5/3, 5/3, 5/3, 0],
                           [0, 0,   0,   0,   0]], dtype=np.float64)
    y = stack.mean_projection(x, return_float=True)
    assert_array_almost_equal(y, expected_y)
    assert y.dtype in [np.float32, np.float64]


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.float32, np.float64])
def test_median_projection(dtype):
    x = x_3d.astype(dtype)
    expected_y = np.array([[2, 0, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 0, 0]], dtype=dtype)
    y = stack.median_projection(x)
    assert_array_equal(y, expected_y)
    assert y.dtype == dtype


# ### test focus selection ###

@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64])
def test_one_hot_3d(dtype):
    # integer
    indices = np.array(
        [[1, 0, 0, 0, 0],
         [0, 1, 0, 0, 0],
         [2, 1, 0, 2, 2],
         [0, 1, 1, 1, 2],
         [0, 0, 0, 1, 0]], dtype=dtype)
    one_hot = _one_hot_3d(indices=indices, depth=4)
    expected_one_hot = np.array(
        [[[0, 1, 1, 1, 1],
          [1, 0, 1, 1, 1],
          [0, 0, 1, 0, 0],
          [1, 0, 0, 0, 0],
          [1, 1, 1, 0, 1]],

         [[1, 0, 0, 0, 0],
          [0, 1, 0, 0, 0],
          [0, 1, 0, 0, 0],
          [0, 1, 1, 1, 0],
          [0, 0, 0, 1, 0]],

         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [1, 0, 0, 1, 1],
          [0, 0, 0, 0, 1],
          [0, 0, 0, 0, 0]],

         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0]]],
        dtype=np.uint8)
    assert_array_equal(one_hot, expected_one_hot)
    assert one_hot.dtype == indices.dtype

    # boolean
    one_hot = _one_hot_3d(indices=indices, depth=4, return_boolean=True)
    expected_one_hot = expected_one_hot.astype(bool)
    assert_array_equal(one_hot, expected_one_hot)
    assert one_hot.dtype == bool


def test_get_in_focus_indices():
    focus = stack.compute_focus(x_3d_out_focus, neighborhood_size=31)

    # number of slices to keep
    indices_to_keep = stack.get_in_focus_indices(focus, proportion=3)
    assert isinstance(indices_to_keep, list)
    assert len(indices_to_keep) == 3
    assert indices_to_keep[0] == 1
    assert indices_to_keep[1] == 3
    assert indices_to_keep[2] == 4

    # proportion of slices to keep
    indices_to_keep = stack.get_in_focus_indices(focus, proportion=0.5)
    assert isinstance(indices_to_keep, list)
    assert len(indices_to_keep) == 3
    assert indices_to_keep[0] == 1
    assert indices_to_keep[1] == 3
    assert indices_to_keep[2] == 4

    # 0 slice
    indices_to_keep = stack.get_in_focus_indices(focus, proportion=0)
    assert isinstance(indices_to_keep, list)
    assert len(indices_to_keep) == 0

    # error: float > 1
    with pytest.raises(ValueError):
        _ = stack.get_in_focus_indices(focus, proportion=3.)

    # error: 'proportion' < 0
    with pytest.raises(ValueError):
        _ = stack.get_in_focus_indices(focus, proportion=-3)


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.float32, np.float64])
def test_in_focus_selection(dtype):
    x = x_3d_out_focus.astype(dtype)
    expected_y = x_3d.astype(dtype)
    focus = stack.compute_focus(x_3d_out_focus, neighborhood_size=31)
    y = stack.in_focus_selection(x, focus, proportion=3)
    assert_array_equal(y, expected_y)
    assert y.dtype == dtype


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.float32, np.float64])
def test_focus_projection(dtype):
    x = x_3d_out_focus.astype(dtype)

    # 'method' == "median"
    expected_y = np.array([[2, 0, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 0, 0]], dtype=dtype)
    y = stack.focus_projection(
        x, proportion=3, neighborhood_size=7, method="median")
    assert_array_equal(y, expected_y)
    assert y.dtype == dtype

    # 'method' == "max"
    expected_y = np.array([[3, 0, 0, 0, 0],
                           [0, 3, 0, 0, 3],
                           [0, 3, 0, 0, 0],
                           [0, 3, 3, 3, 0],
                           [0, 0, 0, 0, 0]], dtype=dtype)
    y = stack.focus_projection(
        x, proportion=3, neighborhood_size=7, method="max")
    assert_array_equal(y, expected_y)
    assert y.dtype == dtype

    # error: method not in ["median", "max"]
    with pytest.raises(ValueError):
        _ = stack.focus_projection(
            x, proportion=3, neighborhood_size=7, method="mean")



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions used to detect and clean noisy images.
"""

import numpy as np

from .utils import check_array, check_parameter
from .filter import mean_filter


# ### Focus ###

def compute_focus(image, neighborhood_size=31):
    """Helmli and Scherer’s mean method is used as a focus metric.

    For each pixel yx in a 2-d image, we compute the ratio:

    .. math::

        R(y, x) = \\left \\{ \\begin{array}{rcl} \\frac{I(y, x)}{\\mu(y, x)} &
        \\mbox{if} & I(y, x) \\ge \\mu(y, x) \\\ \\frac{\\mu(y, x)}{I(y, x)} &
        \\mbox{otherwise} & \\end{array} \\right.

    with :math:`I(y, x)` the intensity of the pixel yx and :math:`\\mu(y, x)`
    the mean intensity of the pixels in its neighborhood.

    For a 3-d image, we compute this metric for each z surface.

    Parameters
    ----------
    image : np.ndarray
        A 2-d or 3-d image with shape (y, x) or (z, y, x).
    neighborhood_size : int or tuple or list, default=31
        The size of the square used to define the neighborhood of each pixel.
        An odd value is preferred. To define a rectangular neighborhood, a
        tuple or a list with two elements (height, width) can be provided.

    Returns
    -------
    focus : np.ndarray, np.float64
        A 2-d or 3-d tensor with the R(y, x) computed for each pixel of the
        original image.

    """
    # check parameters
    check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    check_parameter(neighborhood_size=(int, tuple, list))
    if (isinstance(neighborhood_size, (tuple, list))
            and len(neighborhood_size) != 2):
        raise ValueError("Parameter 'neighborhood_size' should be an integer "
                         "(to define a square neighborhood) or a sequence of "
                         "two elements (to define a rectangular "
                         "neighborhood). Not a sequence with {0} elements."
                         .format(len(neighborhood_size)))

    # cast image in float if necessary
    if image.dtype in [np.uint8, np.uint16]:
        image_float = image.astype(np.float64)
    else:
        image_float = image

    # build kernel
    if image.ndim == 3:
        focus = _compute_focus_3d(image_float, neighborhood_size)
    else:
        focus = _compute_focus_2d(image_float, neighborhood_size)

    return focus


def _compute_focus_3d(image_3d, kernel_size):
    """Compute a pixel-wise focus metric for a 3-d image.

    Parameters
    ----------
    image_3d : np.ndarray, np.float
        A 3-d image with shape (z, y, x).
    kernel_size : int or tuple or list
        The size of the square used to define a kernel size. An odd value is
        preferred. To define a rectangular kernel, a tuple or a list with two
        elements (height, width) can be provided.

    Returns
    -------
    focus : np.ndarray, np.float64
        A 3-d tensor with the R(z, y, x) computed for each pixel of the
        original image.

    """
    # compute focus metric for each z surface
    focus = []
    for image_2d in image_3d:
        focus_2d = _compute_focus_2d(image_2d, kernel_size)
        focus.append(focus_2d)

    #  stack focus metrics
    focus = np.stack(focus)

    return focus


def _compute_focus_2d(image_2d, kernel_size):
    """Compute a pixel-wise focus metric for a 2-d image.

    Parameters
    ----------
    image_2d : np.ndarray, np.float
        A 2-d image with shape (y, x).
    kernel_size : int or tuple or list
        The size of the square used to define a kernel size. An odd value is
        preferred. To define a rectangular kernel, a tuple or a list with two
        elements (height, width) can be provided.

    Returns
    -------
    focus : np.ndarray, np.float64
        A 2-d tensor with the R(y, x) computed for each pixel of the original
        image.

    """
    # mean filtered image
    if isinstance(kernel_size, int):
        image_filtered_mean = mean_filter(image_2d, "square", kernel_size)
    else:
        image_filtered_mean = mean_filter(image_2d, "rectangle", kernel_size)

    # compute focus metric
    ratio_default_1 = np.ones_like(image_2d, dtype=np.float64)
    ratio_default_2 = np.ones_like(image_filtered_mean, dtype=np.float64)
    ratio_1 = np.divide(
        image_2d, image_filtered_mean,
        out=ratio_default_1,
        where=image_filtered_mean > 0)
    ratio_2 = np.divide(
        image_filtered_mean, image_2d,
        out=ratio_default_2,
        where=image_2d > 0)
    focus = np.where(image_2d >= image_filtered_mean, ratio_1, ratio_2)

    return focus



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.stack.preprocess module.
"""

import pytest

import numpy as np
import bigfish.stack as stack

from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal


# ### Test normalization ###

@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32,
    np.int8, np.int16, np.int32,
    np.float32, np.float64])
def test_rescale(dtype):
    # build a 5x5 random matrix with a limited range of values
    tensor = np.random.randint(35, 105, 25).reshape((5, 5)).astype(dtype)

    # rescale tensor
    rescaled_tensor = stack.rescale(tensor)

    # test consistency of the function
    if dtype in [np.uint8, np.uint16, np.uint32, np.int8, np.int16, np.int32]:
        i = np.iinfo(dtype)
        min_, max_ = 0, i.max
    else:
        min_, max_ = 0, 1
    assert rescaled_tensor.min() == min_
    assert rescaled_tensor.max() == max_
    assert rescaled_tensor.dtype == dtype
    assert rescaled_tensor.shape == (5, 5)


def test_stretching():
    x = [[51, 51, 51], [102, 102, 102], [153, 153, 153]]

    # integer
    tensor = np.array(x).reshape((3, 3)).astype(np.uint16)
    tensor_rescaled = stack.rescale(tensor,
                                    channel_to_stretch=0,
                                    stretching_percentile=50)
    expected_tensor = np.array([[0, 0, 0],
                                [65535, 65535, 65535],
                                [65535, 65535, 65535]], dtype=np.uint16)
    assert_array_equal(tensor_rescaled, expected_tensor)

    # float
    tensor = np.array(x).reshape((3, 3)).astype(np.float32)
    rescaled_tensor = stack.rescale(tensor,
                                    channel_to_stretch=0,
                                    stretching_percentile=50)
    expected_tensor = np.array([[0., 0., 0.],
                                [1., 1., 1.],
                                [1., 1., 1.]], dtype=np.float32)
    assert_array_equal(rescaled_tensor, expected_tensor)


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float32, np.float64])
def test_cast_uint8(dtype):
    # from integer to np.uint8
    if np.issubdtype(dtype, np.integer):
        x = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)
        tensor[2, 2] = np.iinfo(dtype).max

    # from float to np.uint8
    else:
        x = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 1.0]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)

    # cast in uint8
    tensor_uint8 = stack.cast_img_uint8(tensor)
    assert tensor_uint8.dtype == np.uint8

    # check value
    assert tensor_uint8.max() == 255
    if dtype == np.uint8:
        assert_array_equal(tensor_uint8, tensor)


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float32, np.float64])
def test_cast_uint16(dtype):
    # from integer to np.uint16
    if np.issubdtype(dtype, np.integer):
        x = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)
        tensor[2, 2] = np.iinfo(dtype).max

    # from float to np.uint16
    else:
        x = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 1.0]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)

    # cast in uint16
    tensor_uint16 = stack.cast_img_uint16(tensor)
    assert tensor_uint16.dtype == np.uint16

    # check value
    assert tensor_uint16.max() == 65535
    if dtype == np.uint16:
        assert_array_equal(tensor_uint16, tensor)


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float32, np.float64])
def test_cast_float32(dtype):
    # from integer to np.float32
    if np.issubdtype(dtype, np.integer):
        x = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)
        tensor[0, 0] = np.iinfo(dtype).min
        tensor[2, 2] = np.iinfo(dtype).max

    # from float to np.float32
    else:
        x = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 1.0]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)

    # cast in float32
    tensor_float32 = stack.cast_img_float32(tensor)
    assert tensor_float32.dtype == np.float32

    # check value
    if dtype in [np.uint8, np.uint16, np.uint32, np.uint64]:
        assert tensor_float32.min() == 0
        assert tensor_float32.max() == 1
    elif dtype in [np.int8, np.int16, np.int32, np.int64]:
        assert tensor_float32.min() == -1
        assert tensor_float32.max() == 1
    else:
        assert_array_almost_equal(tensor_float32, tensor)


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float32, np.float64])
def test_cast_float64(dtype):
    # from integer to np.float64
    if np.issubdtype(dtype, np.integer):
        x = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)
        tensor[0, 0] = np.iinfo(dtype).min
        tensor[2, 2] = np.iinfo(dtype).max

    # from float to np.float64
    else:
        x = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 1.0]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)

    # cast in float64
    tensor_float64 = stack.cast_img_float64(tensor)
    assert tensor_float64.dtype == np.float64

    # check value
    if dtype in [np.uint8, np.uint16, np.uint32, np.uint64]:
        assert tensor_float64.min() == 0
        assert tensor_float64.max() == 1
    elif dtype in [np.int8, np.int16, np.int32, np.int64]:
        assert tensor_float64.min() == -1
        assert tensor_float64.max() == 1
    else:
        assert_array_almost_equal(tensor_float64, tensor)



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.stack.utils module.
"""

import os
import pytest
import tempfile

import apifish.multistack as multistack


# ### Test recipes ###

def test_check_recipe():
    # build a temporary directory with two files
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = os.path.join(tmp_dir, "experience_1_dapi_fov_1.tif")
        with open(path, 'w') as f:
            f.write("dapi file")
        path = os.path.join(tmp_dir, "experience_1_smfish_fov_1.tif")
        with open(path, 'w') as f:
            f.write("smFISH file")

        # test the consistency of the check function when it should work
        good_recipe_1 = {"fov": "fov_1",
                         "c": ["dapi", "smfish"],
                         "opt": "experience_1",
                         "ext": "tif",
                         "pattern": "opt_c_fov.ext"}
        assert multistack.check_recipe(good_recipe_1, data_directory=None)
        assert multistack.check_recipe(good_recipe_1, data_directory=tmp_dir)

        # case with a good recipe but when a file is missing
        good_recipe_2 = {"fov": "fov_1",
                         "c": ["dapi", "smfish", "cellmask"],
                         "opt": "experience_1",
                         "ext": "tif",
                         "pattern": "opt_c_fov.ext"}
        assert multistack.check_recipe(good_recipe_2, data_directory=None)
        with pytest.raises(FileNotFoundError):
            multistack.check_recipe(good_recipe_2, data_directory=tmp_dir)

        # cases without a 'pattern' key with a string value
        bad_recipe_1 = {"fov": "fov_1",
                        "c": ["dapi", "smfish"],
                        "opt": "experience_1",
                        "ext": "tif"}
        bad_recipe_2 = {"fov": "fov_1",
                        "c": ["dapi", "smfish"],
                        "opt": "experience_1",
                        "ext": "tif",
                        "pattern": ["opt_c_fov.ext"]}
        with pytest.raises(KeyError):
            multistack.check_recipe(bad_recipe_1, data_directory=None)
        with pytest.raises(TypeError):
            multistack.check_recipe(bad_recipe_2, data_directory=None)

        # case with a wrong pattern (repetitive key)
        bad_recipe_3 = {"fov": "fov_1",
                        "c": ["dapi", "smfish"],
                        "opt": "experience_1",
                        "ext": "tif",
                        "pattern": "opt_c_fov_fov.ext"}
        with pytest.raises(ValueError):
            multistack.check_recipe(bad_recipe_3, data_directory=None)

        # case with wrong key or value
        bad_recipe_4 = {"fov": "fov_1",
                        "channel": ["dapi", "smfish"],
                        "optional": "experience_1",
                        "extension": "tif",
                        "pattern": "opt_c_fov.ext"}
        bad_recipe_5 = {"fov": "fov_1",
                        "c": ["dapi", "smfish"],
                        "opt": 1,
                        "ext": "tif",
                        "pattern": "opt_c_fov.ext"}
        with pytest.raises(KeyError):
            multistack.check_recipe(bad_recipe_4, data_directory=None)
        with pytest.raises(TypeError):
            multistack.check_recipe(bad_recipe_5, data_directory=None)


def test_fit_recipe():
    # build a recipe to fit
    good_recipe = {"fov": "fov_1",
                   "c": ["dapi", "smfish"],
                   "opt": "experience_1",
                   "ext": "tif",
                   "pattern": "opt_c_fov.ext"}

    # fit recipe
    new_recipe = multistack.fit_recipe(good_recipe)

    # all keys should be initialized in the new recipe, with a list or a string
    for key in ['fov', 'r', 'c', 'z']:
        assert key in new_recipe
        assert isinstance(new_recipe[key], list)
    for key in ['ext', 'opt']:
        assert key in new_recipe
        assert isinstance(new_recipe[key], str)
    assert 'pattern' in new_recipe
    assert isinstance(new_recipe['pattern'], str)

    # test that fitting an already fitted recipe does not change anything
    new_recip_bis = multistack.fit_recipe(new_recipe)
    assert new_recip_bis == new_recipe


def test_path_from_recipe():
    # build a temporary directory with one file
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = os.path.join(tmp_dir, "experience_1_dapi_fov_1.tif")
        with open(path, 'w') as f:
            f.write("dapi file")

        # build a recipe to read the file
        good_recipe = {"fov": "fov_1",
                       "c": "dapi",
                       "opt": "experience_1",
                       "ext": "tif",
                       "pattern": "opt_c_fov.ext"}

        # test the path
        path_dapi = multistack.get_path_from_recipe(good_recipe, tmp_dir, c=0)
        assert os.path.isfile(path_dapi)


def test_element_per_dimension():
    # build a recipe to test
    good_recipe = {"fov": "fov_1",
                   "c": ["dapi", "smfish"],
                   "opt": "experience_1",
                   "ext": "tif",
                   "pattern": "opt_c_fov.ext"}

    # test the number of elements to be stacked
    nb_r, nb_c, nb_z = multistack.get_nb_element_per_dimension(good_recipe)
    assert nb_r == 1
    assert nb_c == 2
    assert nb_z == 1


def test_nb_fov():
    # case when 'fov' key is a string
    good_recipe_1 = {"fov": "fov_1",
                     "c": ["dapi", "smfish"],
                     "opt": "experience_1",
                     "ext": "tif",
                     "pattern": "opt_c_fov.ext"}
    nb_fov = multistack.count_nb_fov(good_recipe_1)
    assert nb_fov == 1

    # case when 'fov' key is a list
    good_recipe_2 = {"fov": ["fov_1", "fov_2"],
                     "c": ["dapi", "smfish"],
                     "opt": "experience_1",
                     "ext": "tif",
                     "pattern": "opt_c_fov.ext"}
    nb_fov = multistack.count_nb_fov(good_recipe_2)
    assert nb_fov == 2

    # case when 'fov' key does not exist
    good_recipe_3 = {"c": ["dapi", "smfish"],
                     "opt": "experience_1",
                     "ext": "tif",
                     "pattern": "opt_c_fov.ext"}
    nb_fov = multistack.count_nb_fov(good_recipe_3)
    assert nb_fov == 1

    # case when the 'fov' key is not a string or a list
    with pytest.raises(TypeError):
        bad_recipe = {"fov": 1,
                      "c": ["dapi", "smfish"],
                      "opt": "experience_1",
                      "ext": "tif",
                      "pattern": "opt_c_fov.ext"}
        multistack.count_nb_fov(bad_recipe)


def test_check_datamap():
    # build a temporary directory with two files
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = os.path.join(tmp_dir, "experience_1_dapi_fov_1.tif")
        with open(path, 'w') as f:
            f.write("dapi file")
        path = os.path.join(tmp_dir, "experience_1_smfish_fov_1.tif")
        with open(path, 'w') as f:
            f.write("smFISH file")

        # test the consistency of the check function
        recipe = {"fov": "fov_1",
                  "c": ["dapi", "smfish"],
                  "opt": "experience_1",
                  "ext": "tif",
                  "pattern": "opt_c_fov.ext"}
        datamap = [(recipe, tmp_dir)]
        assert multistack.check_datamap(datamap)
        datamap = [[recipe, tmp_dir]]
        assert multistack.check_datamap(datamap)
        datamap = [(None, tmp_dir)]
        with pytest.raises(TypeError):
            multistack.check_datamap(datamap)
        datamap = [(recipe, 3)]
        with pytest.raises(TypeError):
            multistack.check_datamap(datamap)
        datamap = [(recipe, "/foo/bar")]
        with pytest.raises(NotADirectoryError):
            multistack.check_datamap(datamap)
        datamap = [(recipe, tmp_dir, None)]
        with pytest.raises(ValueError):
            multistack.check_datamap(datamap)



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
The apifish.multistack subpackage includes function to process input and output
from different channels.
"""

from .utils import check_recipe
from .utils import fit_recipe
from .utils import get_path_from_recipe
from .utils import get_nb_element_per_dimension
from .utils import count_nb_fov
from .utils import check_datamap

from .preprocess import build_stacks
from .preprocess import build_stack
from .preprocess import build_stack_no_recipe

from .colocalization import detect_spots_colocalization
from .colocalization import get_elbow_value_colocalized

from .postprocess import identify_objects_in_region
from .postprocess import remove_transcription_site
from .postprocess import match_nuc_cell
from .postprocess import extract_cell
from .postprocess import extract_spots_from_frame
from .postprocess import summarize_extraction_results
from .postprocess import center_mask_coord
from .postprocess import from_boundaries_to_surface
from .postprocess import from_surface_to_boundaries
from .postprocess import from_binary_to_coord
from .postprocess import complete_coord_boundaries
from .postprocess import from_coord_to_frame
from .postprocess import from_coord_to_surface


# TODO complete apifish.multistack.preprocess documentation

_utils = [
    "check_recipe",
    "fit_recipe",
    "get_path_from_recipe",
    "get_nb_element_per_dimension",
    "count_nb_fov",
    "check_datamap"]

_preprocess = [
    "build_stacks",
    "build_stack",
    "build_stack_no_recipe"]

_colocalization = [
    "detect_spots_colocalization",
    "get_elbow_value_colocalized"]

_postprocess = [
    "identify_objects_in_region",
    "remove_transcription_site",
    "match_nuc_cell",
    "extract_cell",
    "extract_spots_from_frame",
    "summarize_extraction_results",
    "center_mask_coord",
    "from_boundaries_to_surface",
    "from_surface_to_boundaries",
    "from_binary_to_coord",
    "complete_coord_boundaries",
    "from_coord_to_frame",
    "from_coord_to_surface"]

__all__ = _utils + _preprocess + _colocalization + _postprocess



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Postprocessing functions functions for apifish.segmentation subpackage.
"""

import apifish.stack as stack

import numpy as np
from scipy import ndimage as ndi

from skimage.measure import label
from skimage.morphology import remove_small_objects


# TODO make functions compatible with different type of integers

# ### Labelled images ###

def label_instances(image_binary):
    """Count and label the different instances previously segmented in an
    image.

    Parameters
    ----------
    image_binary : np.ndarray, bool
        Binary segmented image with shape (z, y, x) or (y, x).

    Returns
    -------
    image_label : np.ndarray, np.int64
        Labelled image. Each instance is characterized by the same pixel value.

    """
    # check parameters
    stack.check_array(image_binary, ndim=[2, 3], dtype=bool)

    # label instances
    image_label = label(image_binary)

    return image_label


def merge_labels(image_label_1, image_label_2):
    """Combine two partial labels of the same image.

    To prevent merging conflict, labels should not be rescale.

    Parameters
    ----------
    image_label_1 : np.ndarray, np.int64
        Labelled image with shape (z, y, x) or (y, x).
    image_label_2 : np.ndarray, np.int64
        Labelled image with shape (z, y, x) or (y, x).

    Returns
    -------
    image_label : np.ndarray, np.int64
        Labelled image with shape (z, y, x) or (y, x).

    """
    # check parameters
    stack.check_array(image_label_1, ndim=[2, 3], dtype=np.int64)
    stack.check_array(image_label_2, ndim=[2, 3], dtype=np.int64)

    # count number of instances
    nb_instances_1 = image_label_1.max()
    nb_instances_2 = image_label_2.max()
    nb_instances = nb_instances_1 + nb_instances_2

    # check if labels can be merged
    if nb_instances > np.iinfo(np.int64).max:
        raise ValueError("Labels can not be merged. There are too many "
                         "instances for a 64 bit image, labels could overlap.")

    # merge labels
    image_label_2[image_label_2 > 0] += image_label_1
    image_label = np.maximum(image_label_1, image_label_2)

    return image_label


# ### Clean segmentation ###
# TODO make it available for 3D images
def clean_segmentation(image, small_object_size=None, fill_holes=False,
                       smoothness=None, delimit_instance=False):
    """Clean segmentation results (binary masks or integer labels).

    Parameters
    ----------
    image : np.ndarray, np.int64 or bool
        Labelled or masked image with shape (y, x).
    small_object_size : int or None
        Areas with a smaller surface (in pixels) are removed.
    fill_holes : bool
        Fill holes within a labelled or masked area.
    smoothness : int or None
        Radius of a median kernel filter. The higher the smoother instance
        boundaries are.
    delimit_instance : bool
        Delimit clearly instances boundaries by preventing contact between each
        others.

    Returns
    -------
    image_cleaned : np.ndarray, np.int64 or bool
        Cleaned image with shape (y, x).

    """
    # check parameters
    stack.check_array(image, ndim=2, dtype=[np.int64, bool])
    stack.check_parameter(
        small_object_size=(int, type(None)),
        fill_holes=bool,
        smoothness=(int, type(None)),
        delimit_instance=bool)

    # initialize cleaned image
    image_cleaned = image.copy()

    # remove small object
    if small_object_size is not None:
        image_cleaned = _remove_small_area(image_cleaned, small_object_size)

    # fill holes
    if fill_holes:
        image_cleaned = _fill_hole(image_cleaned)

    if smoothness:
        image_cleaned = _smooth_instance(image_cleaned, smoothness)

    if delimit_instance:
        image_cleaned = _delimit_instance(image_cleaned)

    return image_cleaned


def _remove_small_area(image, min_size):
    """Remove segmented areas with a small surface.

    Parameters
    ----------
    image : np.ndarray, np.int64 or bool
        Labelled or masked image with shape (y, x).
    min_size : int
        Areas with a smaller surface (in pixels) are removed.

    Returns
    -------
    image_cleaned : np.ndarray, np.int64 or bool
        Cleaned image with shape (y, x).

    """
    # remove small object
    image_cleaned = remove_small_objects(image, min_size=min_size)

    return image_cleaned


def _fill_hole(image):
    """Fill holes within the segmented areas.

    Parameters
    ----------
    image : np.ndarray, np.int64 or bool
        Labelled or masked image with shape (y, x).

    Returns
    -------
    image_cleaned : np.ndarray, np.int64 or bool
        Cleaned image with shape (y, x).

    """
    # fill holes in a binary mask
    if image.dtype == bool:
        image_cleaned = ndi.binary_fill_holes(image)

    # fill holes in a labelled image
    else:
        image_cleaned = np.zeros_like(image)
        for i in range(1, image.max() + 1):
            image_binary = image == i
            image_binary = ndi.binary_fill_holes(image_binary)
            image_cleaned[image_binary] = i

    return image_cleaned


def _smooth_instance(image, radius):
    """Apply a median filter to smooth instance boundaries.

    Parameters
    ----------
    image : np.ndarray, np.int64 or bool
        Labelled or masked image with shape (y, x).
    radius : int
        Radius of the kernel for the median filter. The higher the smoother.

    Returns
    -------
    image_cleaned : np.ndarray, np.int64 or bool
        Cleaned image with shape (y, x).

    """
    # smooth instance boundaries for a binary mask
    if image.dtype == bool:
        image_cleaned = image.astype(np.uint8)
        image_cleaned = stack.median_filter(image_cleaned, "disk", radius)
        image_cleaned = image_cleaned.astype(bool)

    # smooth instance boundaries for a labelled image
    else:
        if image.max() <= 65535 and image.min() >= 0:
            image_cleaned = image.astype(np.uint16)
            image_cleaned = stack.median_filter(image_cleaned, "disk", radius)
            image_cleaned = image_cleaned.astype(np.int64)
        else:
            raise ValueError("Segmentation boundaries can't be smoothed "
                             "because more than 65535 has been detected in "
                             "the image. Smoothing is performed with 16-bit "
                             "unsigned integer images.")

    return image_cleaned


def _delimit_instance(image):
    """Subtract an eroded image to a dilated one in order to prevent
    boundaries contact.

    Parameters
    ----------
    image : np.ndarray, np.int64 or bool
        Labelled or masked image with shape (y, x).

    Returns
    -------
    image_cleaned : np.ndarray, np.int64 or bool
        Cleaned image with shape (y, x).

    """
    # handle 64 bit integer
    original_dtype = image.dtype
    if image.dtype == np.int64:
        image = image.astype(np.float64)

    # erode-dilate mask
    image_dilated = stack.dilation_filter(image, "disk", 1)
    image_eroded = stack.erosion_filter(image, "disk", 1)
    if original_dtype == bool:
        borders = image_dilated & ~image_eroded
        image_cleaned = image.copy()
        image_cleaned[borders] = False
    else:
        borders = image_dilated - image_eroded
        image_cleaned = image.copy()
        image_cleaned[borders > 0] = 0
        image_cleaned = image_cleaned.astype(original_dtype)

    return image_cleaned


def remove_disjoint(image):
    """For each instances with disconnected parts, keep the larger one.

    Parameters
    ----------
    image : np.ndarray, np.int, np.uint or bool
        Labelled image with shape (z, y, x) or (y, x).

    Returns
    -------
    image_cleaned : np.ndarray, np.int or np.uint
        Cleaned image with shape (z, y, x) or (y, x).

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.int64, bool])

    # handle boolean array
    cast_to_bool = False
    if image.dtype == bool:
        cast_to_bool = bool
        image = image.astype(np.uint8)

    # initialize cleaned labels
    image_cleaned = np.zeros_like(image)

    # loop over instances
    max_label = image.max()
    for i in range(1, max_label + 1):

        # get instance mask
        mask = image == i

        # check if an instance is labelled with this value
        if mask.sum() == 0:
            continue

        # get an index for each disconnected part of the instance
        labelled_mask = label(mask)
        indices = sorted(list(set(labelled_mask.ravel())))
        if 0 in indices:
            indices = indices[1:]

        # keep the largest part of the instance
        max_area = 0
        mask_instance = None
        for j in indices:
            mask_part_j = labelled_mask == j
            area_j = mask_part_j.sum()
            if area_j > max_area:
                max_area = area_j
                mask_instance = mask_part_j

        # add instance in the final label
        image_cleaned[mask_instance] = i

    if cast_to_bool:
        image_cleaned = image_cleaned.astype(bool)

    return image_cleaned



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Utility functions for apifish.detection subpackage.
"""

import warnings

import numpy as np

import apifish.stack as stack


# ### Pixel - nanometer conversion

def convert_spot_coordinates(spots, voxel_size):
    """Convert spots coordinates from pixel to nanometer.

    Parameters
    ----------
    spots : np.ndarray
        Coordinates of the detected spots with shape (nb_spots, 3) or
        (nb_spots, 2).
    voxel_size : int, float, Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.

    Returns
    -------
    spots_nanometer : np.ndarray
        Coordinates of the detected spots with shape (nb_spots, 3) or
        (nb_spots, 3), in nanometer.

    """
    # check parameters
    stack.check_parameter(voxel_size=(int, float, tuple, list))
    stack.check_array(spots, ndim=2, dtype=[np.float64, np.int64])
    dtype = spots.dtype

    # check consistency between parameters
    ndim = spots.shape[1]
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError("'voxel_size' must be a scalar or a sequence "
                             "with {0} elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim

    # convert spots coordinates in nanometer
    spots_nanometer = spots.copy()
    if ndim == 3:
        spots_nanometer[:, 0] *= voxel_size[0]
        spots_nanometer[:, 1:] *= voxel_size[-1]

    else:
        spots_nanometer *= voxel_size[-1]
    spots_nanometer = spots_nanometer.astype(dtype)

    return spots_nanometer


def get_object_radius_pixel(voxel_size_nm, object_radius_nm, ndim):
    """Convert the object radius in pixel.

    When the object considered is a spot this value can be interpreted as the
    standard deviation of the spot PSF, in pixel. For any object modelled with
    a gaussian signal, this value can be interpreted as the standard deviation
    of the gaussian.

    Parameters
    ----------
    voxel_size_nm : int, float, Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    object_radius_nm : int, float, Tuple(int, float) or List(int, float)
        Radius of the object, in nanometer. One value per spatial dimension
        (zyx or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions.
    ndim : int
        Number of spatial dimension to consider.

    Returns
    -------
    object_radius_px : Tuple[float]
        Radius of the object in pixel, one element per dimension (zyx or yx
        dimensions).

    """
    # check parameters
    stack.check_parameter(
        voxel_size_nm=(int, float, tuple, list),
        object_radius_nm=(int, float, tuple, list),
        ndim=int)

    # check consistency between parameters
    if isinstance(voxel_size_nm, (tuple, list)):
        if len(voxel_size_nm) != ndim:
            raise ValueError("'voxel_size_nm' must be a scalar or a sequence "
                             "with {0} elements.".format(ndim))
    else:
        voxel_size_nm = (voxel_size_nm,) * ndim
    if isinstance(object_radius_nm, (tuple, list)):
        if len(object_radius_nm) != ndim:
            raise ValueError("'object_radius_nm' must be a scalar or a "
                             "sequence with {0} elements.".format(ndim))
    else:
        object_radius_nm = (object_radius_nm,) * ndim

    # get radius in pixel
    object_radius_px = [b / a for a, b in zip(voxel_size_nm, object_radius_nm)]
    object_radius_px = tuple(object_radius_px)

    return object_radius_px


def get_object_radius_nm(voxel_size_nm, object_radius_px, ndim):
    """Convert the object radius in nanometer.

    When the object considered is a spot this value can be interpreted as the
    standard deviation of the spot PSF, in nanometer. For any object modelled
    with a gaussian signal, this value can be interpreted as the standard
    deviation of the gaussian.

    Parameters
    ----------
    voxel_size_nm : int, float, Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    object_radius_px : int, float, Tuple(int, float) or List(int, float)
        Radius of the object, in pixel. One value per spatial dimension
        (zyx or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions.
    ndim : int
        Number of spatial dimension to consider.

    Returns
    -------
    object_radius_nm : Tuple[float]
        Radius of the object in nanometer, one element per dimension (zyx or yx
        dimensions).

    """
    # check parameters
    stack.check_parameter(
        voxel_size_nm=(int, float, tuple, list),
        object_radius_px=(int, float, tuple, list),
        ndim=int)

    # check consistency between parameters
    if isinstance(voxel_size_nm, (tuple, list)):
        if len(voxel_size_nm) != ndim:
            raise ValueError("'voxel_size_nm' must be a scalar or a sequence "
                             "with {0} elements.".format(ndim))
    else:
        voxel_size_nm = (voxel_size_nm,) * ndim
    if isinstance(object_radius_px, (tuple, list)):
        if len(object_radius_px) != ndim:
            raise ValueError("'object_radius_px' must be a scalar or a "
                             "sequence with {0} elements.".format(ndim))
    else:
        object_radius_px = (object_radius_px,) * ndim

    # get radius in pixel
    object_radius_nm = [a * b for a, b in zip(voxel_size_nm, object_radius_px)]
    object_radius_nm = tuple(object_radius_nm)

    return object_radius_nm


# ### Reference spot ###

def build_reference_spot(image, spots, voxel_size, spot_radius, alpha=0.5):
    """Build a median or mean spot in 3 or 2 dimensions as reference.

    Reference spot is computed from a sample of uncropped detected spots. If
    such sample is not possible, an empty frame is returned.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    spots : np.ndarray
        Coordinate of the spots with shape (nb_spots, 3) for 3-d images or
        (nb_spots, 2) for 2-d images.
    voxel_size : int, float, Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    spot_radius : int, float, Tuple(int, float) or List(int, float)
        Radius of the spot, in nanometer. One value per spatial dimension (zyx
        or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions.
    alpha : int or float
        Intensity score of the reference spot, between 0 and 1. If 0, reference
        spot approximates the spot with the lowest intensity. If 1, reference
        spot approximates the brightest spot. Default is 0.5.

    Returns
    -------
    reference_spot : np.ndarray
        Reference spot in 3-d or 2-d.

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_array(spots, ndim=2, dtype=[np.float64, np.int64])
    stack.check_parameter(
        voxel_size=(int, float, tuple, list),
        spot_radius=(int, float, tuple, list),
        alpha=(int, float))
    if alpha < 0 or alpha > 1:
        raise ValueError("'alpha' should be a value between 0 and 1, not {0}"
                         .format(alpha))

    # check consistency between parameters
    ndim = image.ndim
    if ndim != spots.shape[1]:
        raise ValueError("Provided image has {0} dimensions but spots are "
                         "detected in {1} dimensions."
                         .format(ndim, spots.shape[1]))
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError(
                "'voxel_size' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim
    if isinstance(spot_radius, (tuple, list)):
        if len(spot_radius) != ndim:
            raise ValueError(
                "'spot_radius' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        spot_radius = (spot_radius,) * ndim

    # compute radius used to crop spot image
    radius_pixel = get_object_radius_pixel(
        voxel_size_nm=voxel_size,
        object_radius_nm=spot_radius,
        ndim=ndim)
    radius = [np.sqrt(ndim) * r for r in radius_pixel]
    radius = tuple(radius)

    # build reference spot
    if image.ndim == 3:
        reference_spot = _build_reference_spot_3d(image, spots, radius, alpha)
    else:
        reference_spot = _build_reference_spot_2d(image, spots, radius, alpha)

    return reference_spot


def _build_reference_spot_3d(image, spots, radius, alpha):
    """Build a median or mean spot in 3 dimensions as reference.

    Reference spot is computed from a sample of uncropped detected spots. If
    such sample is not possible, an empty frame is returned.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x).
    spots : np.ndarray, np.int64
        Coordinate of the spots with shape (nb_spots, 3) for 3-d images.
    radius : Tuple[float]
        Radius in pixel of the detected spots, one element per dimension.
    alpha : int or float
        Intensity score of the reference spot, between 0 and 1. If 0, reference
        spot approximates the spot with the lowest intensity. If 1, reference
        spot approximates the brightest spot.

    Returns
    -------
    reference_spot : np.ndarray
        Reference spot in 3-d.

    """
    # get a rounded radius for each dimension
    radius_z = np.ceil(radius[0]).astype(np.int64)
    z_shape = radius_z * 2 + 1
    radius_yx = np.ceil(radius[-1]).astype(np.int64)
    yx_shape = radius_yx * 2 + 1

    # randomly choose some spots to aggregate
    indices = [i for i in range(spots.shape[0])]
    np.random.shuffle(indices)
    indices = indices[:min(2000, spots.shape[0])]
    candidate_spots = spots[indices, :]

    # collect area around each spot
    l_reference_spot = []
    for i_spot in range(candidate_spots.shape[0]):

        # get spot coordinates
        spot_z, spot_y, spot_x = candidate_spots[i_spot, :]

        # get the volume of the spot
        image_spot, _, = _get_spot_volume(image, spot_z, spot_y, spot_x,
                                          radius_z, radius_yx)

        # keep images that are not cropped by the borders
        if image_spot.shape == (z_shape, yx_shape, yx_shape):
            l_reference_spot.append(image_spot)

    # if not enough spots are detected
    if len(l_reference_spot) <= 30:
        warnings.warn("Problem occurs during the computation of a reference "
                      "spot. Not enough (uncropped) spots have been detected.",
                      UserWarning)
    if len(l_reference_spot) == 0:
        reference_spot = np.zeros(
            (z_shape, yx_shape, yx_shape), dtype=image.dtype)
        return reference_spot

    # project the different spot images
    l_reference_spot = np.stack(l_reference_spot, axis=0)
    alpha_ = alpha * 100
    reference_spot = np.percentile(l_reference_spot, alpha_, axis=0)
    reference_spot = reference_spot.astype(image.dtype)

    return reference_spot


def _get_spot_volume(image, spot_z, spot_y, spot_x, radius_z, radius_yx):
    """Get a subimage of a detected spot in 3 dimensions.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x).
    spot_z : np.int64
        Coordinate of the detected spot along the z axis.
    spot_y : np.int64
        Coordinate of the detected spot along the y axis.
    spot_x : np.int64
        Coordinate of the detected spot along the x axis.
    radius_z : int
        Radius in pixel of the detected spot, along the z axis.
    radius_yx : int
        Radius in pixel of the detected spot, on the yx plan.

    Returns
    -------
    image_spot : np.ndarray
        Reference spot in 3-d.
    _ : Tuple[int]
        Lower zyx coordinates of the crop.

    """
    # get boundaries of the volume surrounding the spot
    z_spot_min = max(0, int(spot_z - radius_z))
    z_spot_max = min(image.shape[0], int(spot_z + radius_z))
    y_spot_min = max(0, int(spot_y - radius_yx))
    y_spot_max = min(image.shape[1], int(spot_y + radius_yx))
    x_spot_min = max(0, int(spot_x - radius_yx))
    x_spot_max = min(image.shape[2], int(spot_x + radius_yx))

    # get the volume of the spot
    image_spot = image[z_spot_min:z_spot_max + 1,
                       y_spot_min:y_spot_max + 1,
                       x_spot_min:x_spot_max + 1]

    return image_spot, (z_spot_min, y_spot_min, x_spot_min)


def _build_reference_spot_2d(image, spots, radius, alpha):
    """Build a median or mean spot in 2 dimensions as reference.

    Reference spot is computed from a sample of uncropped detected spots. If
    such sample is not possible, an empty frame is returned.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (y, x).
    spots : np.ndarray, np.int64
        Coordinate of the spots with shape (nb_spots, 2) for 2-d images.
    radius : Tuple[float]
        Radius in pixel of the detected spots, one element per dimension.
    alpha : int or float
        Intensity score of the reference spot, between 0 and 1. If 0, reference
        spot approximates the spot with the lowest intensity. If 1, reference
        spot approximates the brightest spot.

    Returns
    -------
    reference_spot : np.ndarray
        Reference spot in 2-d.

    """
    # get a rounded radius for each dimension
    radius_yx = np.ceil(radius[-1]).astype(np.int64)
    yx_shape = radius_yx * 2 + 1

    # randomly choose some spots to aggregate
    indices = [i for i in range(spots.shape[0])]
    np.random.shuffle(indices)
    indices = indices[:min(2000, spots.shape[0])]
    candidate_spots = spots[indices, :]

    # collect area around each spot
    l_reference_spot = []
    for i_spot in range(candidate_spots.shape[0]):

        # get spot coordinates
        spot_y, spot_x = candidate_spots[i_spot, :]

        # get the volume of the spot
        image_spot, _ = _get_spot_surface(image, spot_y, spot_x, radius_yx)

        # keep images that are not cropped by the borders
        if image_spot.shape == (yx_shape, yx_shape):
            l_reference_spot.append(image_spot)

    # if not enough spots are detected
    if len(l_reference_spot) <= 30:
        warnings.warn("Problem occurs during the computation of a reference "
                      "spot. Not enough (uncropped) spots have been detected.",
                      UserWarning)
    if len(l_reference_spot) == 0:
        reference_spot = np.zeros((yx_shape, yx_shape), dtype=image.dtype)
        return reference_spot

    # project the different spot images
    l_reference_spot = np.stack(l_reference_spot, axis=0)
    alpha_ = alpha * 100
    reference_spot = np.percentile(l_reference_spot, alpha_, axis=0)
    reference_spot = reference_spot.astype(image.dtype)

    return reference_spot


def _get_spot_surface(image, spot_y, spot_x, radius_yx):
    """Get a subimage of a detected spot in 2 dimensions.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (y, x).
    spot_y : np.int64
        Coordinate of the detected spot along the y axis.
    spot_x : np.int64
        Coordinate of the detected spot along the x axis.
    radius_yx : int
        Radius in pixel of the detected spot, on the yx plan.

    Returns
    -------
    image_spot : np.ndarray
        Reference spot in 2-d.
    _ : Tuple[int]
        Lower yx coordinates of the crop.

    """
    # get boundaries of the surface surrounding the spot
    y_spot_min = max(0, int(spot_y - radius_yx))
    y_spot_max = min(image.shape[0], int(spot_y + radius_yx))
    x_spot_min = max(0, int(spot_x - radius_yx))
    x_spot_max = min(image.shape[1], int(spot_x + radius_yx))

    # get the surface of the spot
    image_spot = image[y_spot_min:y_spot_max + 1,
                       x_spot_min:x_spot_max + 1]

    return image_spot, (y_spot_min, x_spot_min)


# ### SNR ###

def compute_snr_spots(image, spots, voxel_size, spot_radius):
    """Compute signal-to-noise ratio (SNR) based on spot coordinates.

    .. math::

        \\mbox{SNR} = \\frac{\\mbox{max(spot signal)} -
        \\mbox{mean(background)}}{\\mbox{std(background)}}

    Background is a region twice larger surrounding the spot region. Only the
    y and x dimensions are taking into account to compute the SNR.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    spots : np.ndarray
        Coordinate of the spots, with shape (nb_spots, 3) or (nb_spots, 2).
        One coordinate per dimension (zyx or yx coordinates).
    voxel_size : int, float, Tuple(int, float), List(int, float) or None
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions. Not used if 'log_kernel_size' and 'minimum_distance' are
        provided.
    spot_radius : int, float, Tuple(int, float), List(int, float) or None
        Radius of the spot, in nanometer. One value per spatial dimension (zyx
        or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions. Not used if 'log_kernel_size' and 'minimum_distance'
        are provided.

    Returns
    -------
    snr : float
        Median signal-to-noise ratio computed for every spots.

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_range_value(image, min_=0)
    stack.check_array(spots, ndim=2, dtype=[np.float64, np.int64])
    stack.check_parameter(
        voxel_size=(int, float, tuple, list),
        spot_radius=(int, float, tuple, list))

    # check consistency between parameters
    ndim = image.ndim
    if ndim != spots.shape[1]:
        raise ValueError("Provided image has {0} dimensions but spots are "
                         "detected in {1} dimensions."
                         .format(ndim, spots.shape[1]))
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError(
                "'voxel_size' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim
    if isinstance(spot_radius, (tuple, list)):
        if len(spot_radius) != ndim:
            raise ValueError(
                "'spot_radius' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        spot_radius = (spot_radius,) * ndim

    # cast spots coordinates if needed
    if spots.dtype == np.float64:
        spots = np.round(spots).astype(np.int64)

    # cast image if needed
    image_to_process = image.copy().astype(np.float64)

    # clip coordinate if needed
    if ndim == 3:
        spots[:, 0] = np.clip(spots[:, 0], 0, image_to_process.shape[0] - 1)
        spots[:, 1] = np.clip(spots[:, 1], 0, image_to_process.shape[1] - 1)
        spots[:, 2] = np.clip(spots[:, 2], 0, image_to_process.shape[2] - 1)
    else:
        spots[:, 0] = np.clip(spots[:, 0], 0, image_to_process.shape[0] - 1)
        spots[:, 1] = np.clip(spots[:, 1], 0, image_to_process.shape[1] - 1)

    # compute radius used to crop spot image
    radius_pixel = get_object_radius_pixel(
        voxel_size_nm=voxel_size,
        object_radius_nm=spot_radius,
        ndim=ndim)
    radius_signal_ = [np.sqrt(ndim) * r for r in radius_pixel]
    radius_signal_ = tuple(radius_signal_)

    # compute the neighbourhood radius
    radius_background_ = tuple(i * 2 for i in radius_signal_)

    # ceil radii
    radius_signal = np.ceil(radius_signal_).astype(np.int)
    radius_background = np.ceil(radius_background_).astype(np.int)

    # loop over spots
    snr_spots = []
    for spot in spots:

        # extract spot images
        spot_y = spot[ndim - 2]
        spot_x = spot[ndim - 1]
        radius_signal_yx = radius_signal[-1]
        radius_background_yx = radius_background[-1]
        edge_background_yx = radius_background_yx - radius_signal_yx
        if ndim == 3:
            spot_z = spot[0]
            radius_background_z = radius_background[0]
            max_signal = image_to_process[spot_z, spot_y, spot_x]
            spot_background_, _ = _get_spot_volume(
                image_to_process, spot_z, spot_y, spot_x,
                radius_background_z, radius_background_yx)
            spot_background = spot_background_.copy()

            # discard spot if cropped at the border (along y and x dimensions)
            expected_size = (2 * radius_background_yx + 1) ** 2
            actual_size = spot_background.shape[1] * spot_background.shape[2]
            if expected_size != actual_size:
                continue

            # remove signal from background crop
            spot_background[:,
                            edge_background_yx:-edge_background_yx,
                            edge_background_yx:-edge_background_yx] = -1
            spot_background = spot_background[spot_background >= 0]

        else:
            max_signal = image_to_process[spot_y, spot_x]
            spot_background_, _ = _get_spot_surface(
                image_to_process, spot_y, spot_x, radius_background_yx)
            spot_background = spot_background_.copy()

            # discard spot if cropped at the border
            expected_size = (2 * radius_background_yx + 1) ** 2
            if expected_size != spot_background.size:
                continue

            # remove signal from background crop
            spot_background[edge_background_yx:-edge_background_yx,
                            edge_background_yx:-edge_background_yx] = -1
            spot_background = spot_background[spot_background >= 0]

        # compute mean background
        mean_background = np.mean(spot_background)

        # compute standard deviation background
        std_background = np.std(spot_background)

        # compute SNR
        snr = (max_signal - mean_background) / std_background
        snr_spots.append(snr)

    #  average SNR
    if len(snr_spots) == 0:
        snr = 0.
    else:
        snr = np.median(snr_spots)

    return snr


# ### Miscellaneous ###

def get_breaking_point(x, y):
    """Select the x-axis value where a L-curve has a kink.

    Assuming a L-curve from A to B, the 'breaking_point' is the more distant
    point to the segment [A, B].

    Parameters
    ----------
    x : np.array
        X-axis values.
    y : np.array
        Y-axis values.

    Returns
    -------
    breaking_point : float
        X-axis value at the kink location.
    x : np.array
        X-axis values.
    y : np.array
        Y-axis values.

    """
    # check parameters
    stack.check_array(x, ndim=1, dtype=[np.float64, np.int64])
    stack.check_array(y, ndim=1, dtype=[np.float64, np.int64])

    # select threshold where curve break
    slope = (y[-1] - y[0]) / len(y)
    y_grad = np.gradient(y)
    m = list(y_grad >= slope)
    j = m.index(False)
    m = m[j:]
    x = x[j:]
    y = y[j:]
    if True in m:
        i = m.index(True)
    else:
        i = -1
    breaking_point = float(x[i])

    return breaking_point, x, y



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.classification.features module.
"""

# TODO add test apifish.classification.compute_features
# TODO add test apifish.classification.get_features_name
# TODO add test apifish.classification.features_distance
# TODO add test apifish.classification.features_in_out_nucleus
# TODO add test apifish.classification.features_protrusion
# TODO add test apifish.classification.features_dispersion
# TODO add test apifish.classification.features_topography
# TODO add test apifish.classification.features_foci
# TODO add test apifish.classification.features_area
# TODO add test apifish.classification.features_centrosome



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.multistack.postprocess module.
"""

import pytest

import numpy as np

import apifish.multistack as multistack

from numpy.testing import assert_array_equal


# TODO add test apifish.multistack.match_nuc_cell
# TODO add test apifish.multistack.extract_cell
# TODO add test apifish.multistack.extract_spots_from_frame
# TODO add test apifish.multistack.summarize_extraction_results

@pytest.mark.parametrize("ndim", [2, 3])
@pytest.mark.parametrize("mask_dtype", [
    np.uint8, np.uint16, np.int64, bool])
@pytest.mark.parametrize("spot_dtype", [
    np.int64, np.float64])
def test_identify_objects_in_region(ndim, mask_dtype, spot_dtype):
    # simulate mask and coordinates
    mask = np.zeros((10, 10), dtype=mask_dtype)
    mask[1:4, 1:5] = np.ones((3, 4), dtype=mask_dtype)
    spots_in = [[2, 1, 1], [5, 2, 2], [2, 3, 1], [9, 3, 4]]
    spots_in = np.array(spots_in, dtype=spot_dtype)
    spots_out = [[1, 0, 0], [3, 7, 2], [2, 1, 8], [7, 8, 8]]
    spots_out = np.array(spots_out, dtype=spot_dtype)
    if ndim == 2:
        spots_in = spots_in[:, 1:]
        spots_out = spots_out[:, 1:]
    spots = np.concatenate((spots_in, spots_out))

    # test
    spots_in_, spots_out_ = multistack.identify_objects_in_region(
        mask, spots, ndim)
    assert_array_equal(spots_in_, spots_in)
    assert_array_equal(spots_out_, spots_out)
    assert spots_in_.dtype == spots_in.dtype
    assert spots_out_.dtype == spots_out.dtype


@pytest.mark.parametrize("ndim", [2, 3])
@pytest.mark.parametrize("mask_dtype", [
    np.uint8, np.uint16, np.int64, bool])
@pytest.mark.parametrize("spot_dtype", [
    np.int64, np.float64])
def test_remove_transcription_site(ndim, mask_dtype, spot_dtype):
    # simulate mask and coordinates
    nuc_mask = np.zeros((10, 10), dtype=mask_dtype)
    nuc_mask[1:4, 1:5] = np.ones((3, 4), dtype=mask_dtype)

    spots_in_nuc_out_ts = [[2, 1, 1, -1], [5, 2, 2, -1], [2, 3, 1, -1]]
    spots_in_nuc_out_ts = np.array(spots_in_nuc_out_ts, dtype=spot_dtype)

    spots_in_nuc_in_ts = [[2, 3, 3, 3], [5, 2, 4, 2], [2, 1, 3, 3]]
    spots_in_nuc_in_ts = np.array(spots_in_nuc_in_ts, dtype=spot_dtype)

    spots_out_nuc_out_foci = [[1, 0, 0, -1], [3, 7, 2, -1], [2, 1, 8, -1]]
    spots_out_nuc_out_foci = np.array(spots_out_nuc_out_foci, dtype=spot_dtype)

    spots_out_nuc_in_foci = [[1, 0, 4, 0], [3, 7, 7, 0], [2, 1, 5, 1]]
    spots_out_nuc_in_foci = np.array(spots_out_nuc_in_foci, dtype=spot_dtype)

    if ndim == 2:
        spots_in_nuc_out_ts = spots_in_nuc_out_ts[:, 1:]
        spots_in_nuc_in_ts = spots_in_nuc_in_ts[:, 1:]
        spots_out_nuc_out_foci = spots_out_nuc_out_foci[:, 1:]
        spots_out_nuc_in_foci = spots_out_nuc_in_foci[:, 1:]
    spots = np.concatenate((spots_in_nuc_out_ts, spots_in_nuc_in_ts,
                            spots_out_nuc_out_foci, spots_out_nuc_in_foci))
    spots_out_ts = np.concatenate((spots_in_nuc_out_ts,
                                   spots_out_nuc_out_foci,
                                   spots_out_nuc_in_foci))

    ts = [[2, 2, 4, 1, 2], [4, 2, 3, 2, 3]]
    ts = np.array(ts, dtype=spot_dtype)

    foci = [[0, 6, 7, 2, 0], [1, 2, 5, 1, 1]]
    foci = np.array(foci, dtype=spot_dtype)

    if ndim == 2:
        ts = ts[:, 1:]
        foci = foci[:, 1:]
    all_clusters = np.concatenate((ts, foci))

    # test
    spots_out_ts_, foci_, ts_ = multistack.remove_transcription_site(
        spots, all_clusters, nuc_mask, ndim)
    assert_array_equal(spots_out_ts_, spots_out_ts)
    assert spots_out_ts_.dtype == spots_out_ts.dtype
    assert_array_equal(foci_, foci)
    assert foci_.dtype == foci.dtype
    assert_array_equal(ts_, ts)
    assert ts_.dtype == ts.dtype



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.classification.input_preparation module.
"""

# TODO add test apifish.classification.prepare_extracted_data



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.multistack.colocalization module.
"""

# TODO add test apifish.multistack.detect_spots_colocalization
# TODO add test apifish.multistack.get_elbow_value_colocalized



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.segmentation.utils module.
"""


# TODO add test for apifish.segmentation.thresholding
# TODO add test for apifish.segmentation.compute_mean_diameter
# TODO add test for apifish.segmentation.compute_mean_convexity_ratio
# TODO add test for apifish.segmentation.compute_surface_ratio
# TODO add test for apifish.segmentation.count_instances
# TODO add test for apifish.segmentation.resize_image
# TODO add test for apifish.segmentation.get_marge_padding
# TODO add test for apifish.segmentation.compute_image_standardization



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions to detect colocalized spots in 2-d and 3-d.
"""

import numpy as np

import apifish.stack as stack
import apifish.detection as detection

from scipy.spatial.distance import cdist
from scipy.optimize import linear_sum_assignment
from scipy.signal import savgol_filter


# TODO process multiple images together

# ### Main function ###

def detect_spots_colocalization(spots_1, spots_2, voxel_size, threshold=None,
                                return_indices=False, return_threshold=False):
    """Detect colocalized spots between two arrays of spot coordinates
    'spots_1' and 'spots_2'. Pairs of spots below a specific threshold are
    defined as colocalized.

    Parameters
    ----------
    spots_1 : np.ndarray
        Coordinates of the spots 1 with shape (nb_spots_1, 3) or
        (nb_spots_1, 2).
    spots_2 : np.ndarray
        Coordinates of the spots 2 with shape (nb_spots_2, 3) or
        (nb_spots_2, 2).
    voxel_size : int, float, Tuple(int, float), or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    threshold : int, float or None
        A threshold to discriminate colocalized spots from distant ones. If
        None, an optimal threshold is selected automatically.
    return_indices : bool
        Return the indices of the colocalized spots within 'spots_1' and
        'spots_2'.
    return_threshold : bool
        Return the threshold used to detect colocalized spots.

    Returns
    -------
    spots_1_colocalized : np.ndarray
        Coordinates of the colocalized spots from 'spots_1' with shape
        (nb_colocalized_spots,).
    spots_2_colocalized : np.ndarray
        Coordinates of the colocalized spots from 'spots_2'with shape
        (nb_colocalized_spots,).
    distances : np.ndarray, np.float64
        Distance matrix between spots with shape (nb_colocalized_spots,).
    indices_1 : np.ndarray, np.int64
        Indices of the colocalized spots in 'spots_1' with shape
        (nb_colocalized_spots,). Optional.
    indices_2 : np.ndarray, np.int64
        Indices of the colocalized spots in 'spots_2' with shape
        (nb_colocalized_spots,). Optional.
    threshold : int or float
        Threshold used to discriminate colocalized spots from distant ones.
        Optional.

    """
    # check parameters
    stack.check_parameter(
        voxel_size=(int, float, tuple, list),
        threshold=(float, int, type(None)),
        return_indices=bool,
        return_threshold=bool)

    # check spots coordinates
    stack.check_array(spots_1, ndim=2, dtype=[np.float64, np.int64])
    stack.check_array(spots_2, ndim=2, dtype=[np.float64, np.int64])

    # convert spots coordinates in nanometer
    spots_1_nanometer = detection.convert_spot_coordinates(
        spots=spots_1,
        voxel_size=voxel_size)
    spots_2_nanometer = detection.convert_spot_coordinates(
        spots=spots_2,
        voxel_size=voxel_size)

    # compute distance matrix between spots
    distance_matrix = cdist(spots_1_nanometer, spots_2_nanometer)

    # assign spots based on their euclidean distance
    indices_1, indices_2 = linear_sum_assignment(distance_matrix)

    # get distance between colocalized spots
    distances = distance_matrix[indices_1, indices_2]

    # keep colocalized spots under a specific threshold
    if threshold is None:
        threshold = _automated_threshold_setting_colocalization(distances)

    # keep colocalized spots within a specific distance
    mask = distances <= threshold
    indices_1 = indices_1[mask]
    indices_2 = indices_2[mask]
    distances = distances[mask]

    # get colocalized spots
    spots_1_colocalized = spots_1[indices_1, ...]
    spots_2_colocalized = spots_2[indices_2, ...]

    # return indices and threshold or not
    if return_indices and return_threshold:
        return (spots_1_colocalized, spots_2_colocalized, distances, indices_1,
                indices_2, threshold)
    elif return_indices and not return_threshold:
        return (spots_1_colocalized, spots_2_colocalized, distances, indices_1,
                indices_2)
    elif not return_indices and return_threshold:
        return spots_1_colocalized, spots_2_colocalized, distances, threshold
    else:
        return spots_1_colocalized, spots_2_colocalized, distances


def _automated_threshold_setting_colocalization(distances):
    """Automatically set the optimal threshold to detect colocalized spots
    between two arrays of spot coordinates 'spots_1' and 'spots_2'.

    Parameters
    ----------
    distances : np.ndarray, np.float64
        Distance matrix between spots with shape
        (min(nb_spots_1, nb_spots_2),).

    Returns
    -------
    optimal_threshold : int
        Optimal threshold to discriminate distant spots and colocalized ones.

    """
    # get threshold values we want to test
    min_threshold = distances.min()
    max_threshold = distances.max() + 10
    n_candidates = min(int(max_threshold - min_threshold), 10000)
    thresholds = np.linspace(min_threshold, max_threshold, num=n_candidates)

    # get colocalized spots count
    nb_colocalized = []
    for threshold in thresholds:
        mask = distances <= threshold
        n = mask.sum()
        nb_colocalized.append(n)
    nb_colocalized = np.array(nb_colocalized)

    # select threshold where the break of the distribution is located
    x = thresholds.copy()
    y = -nb_colocalized.copy() + nb_colocalized.max()
    y_smooth = savgol_filter(y, 501, 3, mode="nearest")
    if y_smooth.size > 0:
        optimal_threshold, _, _ = detection.get_breaking_point(x, y_smooth)

    # case where no spots were detected
    else:
        optimal_threshold = None

    return optimal_threshold


def get_elbow_value_colocalized(spots_1, spots_2, voxel_size):
    """Get values to plot the elbow curve used to automatically set the
    threshold to detect colocalized spots.

    Parameters
    ----------
    spots_1 : np.ndarray
        Coordinates of the spots with shape (nb_spots, 3) or (nb_spots, 2).
    spots_2 : np.ndarray
        Coordinates of the spots with shape (nb_spots, 3) or (nb_spots, 2).
    voxel_size : int, float, Tuple(int, float), or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.

    Returns
    -------
    thresholds : np.ndarray, np.float64
        Candidate threshold values.
    nb_colocalized : np.ndarray, np.float64
        Colocalized spots count.
    optimal_threshold : float or None
        Threshold automatically set.

    """
    # check parameters
    stack.check_parameter(voxel_size=(int, float, tuple, list))
    stack.check_array(spots_1, ndim=2, dtype=[np.float64, np.int64])
    stack.check_array(spots_2, ndim=2, dtype=[np.float64, np.int64])

    # check consistency between parameters
    ndim = spots_1.shape[1]
    if ndim not in [2, 3]:
        raise ValueError("Spot coordinates should be in 2 or 3 dimensions, "
                         "not {0}.".format(ndim))
    if spots_2.shape[1] != ndim:
        raise ValueError("Spot coordinates should have the same number of "
                         "dimensions.")
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError(
                "'voxel_size' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim

    # convert spots coordinates in nanometer
    spots_1_nanometer = detection.convert_spot_coordinates(
        spots=spots_1,
        voxel_size=voxel_size)
    spots_2_nanometer = detection.convert_spot_coordinates(
        spots=spots_2,
        voxel_size=voxel_size)

    # compute distance matrix between spots
    distance_matrix = cdist(spots_1_nanometer, spots_2_nanometer)

    # assign spots based on their euclidean distance
    indices_1, indices_2 = linear_sum_assignment(distance_matrix)

    # get distance between colocalized spots
    distances = distance_matrix[indices_1, indices_2]

    # get candidate thresholds
    min_threshold = distances.min()
    max_threshold = distances.max() + 10
    n_candidates = min(int(max_threshold - min_threshold), 10000)
    thresholds = np.linspace(min_threshold, max_threshold, num=n_candidates)

    # get colocalized spots count
    nb_colocalized = []
    for threshold in thresholds:
        mask = distances <= threshold
        n = mask.sum()
        nb_colocalized.append(n)
    nb_colocalized = np.array(nb_colocalized)

    # select threshold where the break of the distribution is located
    x = thresholds.copy()
    y = -nb_colocalized.copy() + nb_colocalized.max()
    y_smooth = savgol_filter(y, 501, 3, mode="nearest")
    optimal_threshold, _, _ = detection.get_breaking_point(x, y_smooth)

    return thresholds, nb_colocalized, optimal_threshold



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Function to plot quality control indicators.
"""

import apifish.stack as stack
import apifish.detection as detection
import apifish.multistack as multistack

import matplotlib.pyplot as plt
import numpy as np

from .utils import save_plot


# ### Focus - sharpness ###

def plot_sharpness(focus_measures, labels=None, title=None, framesize=(5, 5),
                   size_title=20, size_axes=15, size_legend=15,
                   path_output=None, ext="png", show=True):
    """Plot focus measures of a 3-d image, at the z-slice level.

    A measure of focus for each z-slice can be computed by averaging the
    pixel-wise focus measure returned from :func:`apifish.stack.compute_focus`.

    Parameters
    ----------
    focus_measures : np.ndarray or list
        A list of 1-d arrays with the sharpness measure for each z-slices.
    labels : str or list, optional
        List of labels for the different measures to compare.
    title : str, optional
        Title of the plot.
    framesize : tuple, default=(5, 5)
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    size_title : int, default=20
        Size of the title.
    size_axes : int, default=15
        Size of the axes label.
    size_legend : int, default=15
        Size of the legend.
    path_output : str, optional
        Path to save the image (without extension).
    ext : str or list, default='png'
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool, default=True
        Show the figure or not.

    """
    # enlist values if necessary
    if isinstance(focus_measures, np.ndarray):
        focus_measures = [focus_measures]
    if labels is not None and isinstance(labels, str):
        labels = [labels]

    # check parameters
    stack.check_parameter(
        focus_measures=list,
        labels=(list, type(None)),
        title=(str, list, type(None)),
        framesize=tuple,
        size_title=int,
        size_axes=int,
        size_legend=int,
        path_output=(str, type(None)),
        ext=(str, list),
        show=bool)
    length = 0
    for focus_measure in focus_measures:
        stack.check_array(
            focus_measure,
            ndim=1,
            dtype=[np.float32, np.float64])
        length = max(length, focus_measure.size)

    # plot
    plt.figure(figsize=framesize)
    y = np.array([i for i in range(length)])
    for i, focus_measure in enumerate(focus_measures):
        if labels is not None:
            plt.plot(focus_measure, y, label=labels[i])
        else:
            plt.plot(focus_measure, y)

    # axes
    if title is not None:
        plt.title(title, fontweight="bold", fontsize=size_title)
    plt.xlabel("sharpness measure", fontweight="bold", fontsize=size_axes)
    plt.ylabel("z-slices", fontweight="bold", fontsize=size_axes)
    if labels is not None:
        plt.legend(prop={'size': size_legend})

    plt.tight_layout()
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


# ### Elbow plots ###

def plot_elbow(images, voxel_size=None, spot_radius=None, log_kernel_size=None,
               minimum_distance=None, title=None, framesize=(5, 5),
               size_title=20, size_axes=15, size_legend=15, path_output=None,
               ext="png", show=True):
    """Plot the elbow curve that allows an automated spot detection.

    Parameters
    ----------
    images : list
        List of ndarrays with shape (z, y, x) or (y, x). The same threshold is
        applied to every images.
    voxel_size : int or float or tuple or list, optional
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions. Not used if 'log_kernel_size' and 'minimum_distance' are
        provided.
    spot_radius : int or float or tuple or list, optional
        Radius of the spot, in nanometer. One value per spatial dimension (zyx
        or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions. Not used if 'log_kernel_size' and 'minimum_distance'
        are provided.
    log_kernel_size : int or float or tuple or list, optional
        Size of the LoG kernel. It equals the standard deviation (in pixels)
        used for the gaussian kernel (one for each dimension). One value per
        spatial dimension (zyx or yx dimensions). If it's a scalar, the same
        standard deviation is applied to every dimensions. If None, we estimate
        it with the voxel size and spot radius.
    minimum_distance : int or float or tuple or list, optional
        Minimum distance (in pixels) between two spots we want to be able to
        detect separately. One value per spatial dimension (zyx or yx
        dimensions). If it's a scalar, the same distance is applied to every
        dimensions. If None, we estimate it with the voxel size and spot
        radius.
    title : str, optional
        Title of the plot.
    framesize : tuple, default=(5, 5)
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    size_title : int, default=20
        Size of the title.
    size_axes : int, default=15
        Size of the axes label.
    size_legend : int, default=15
        Size of the legend.
    path_output : str, optional
        Path to save the image (without extension).
    ext : str or list, default='png'
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool, default=True
        Show the figure or not.

    """
    # check parameters
    stack.check_parameter(
        title=(str, list, type(None)),
        framesize=tuple,
        size_title=int,
        size_axes=int,
        size_legend=int,
        path_output=(str, type(None)),
        ext=(str, list),
        show=bool)

    # get candidate thresholds and spots count to plot the elbow curve
    thresholds, count_spots, threshold = detection.get_elbow_values(
        images,
        voxel_size=voxel_size,
        spot_radius=spot_radius,
        log_kernel_size=log_kernel_size,
        minimum_distance=minimum_distance)

    # plot
    plt.figure(figsize=framesize)
    plt.plot(thresholds, count_spots, c="#2c7bb6", lw=2)
    if threshold is not None:
        i_threshold = np.argmax(thresholds == threshold)
        plt.scatter(
            threshold,
            count_spots[i_threshold],
            marker="D",
            c="#d7191c",
            s=60,
            label="Selected threshold")

    # axes
    if title is not None:
        plt.title(title, fontweight="bold", fontsize=size_title)
    plt.xlabel("Thresholds", fontweight="bold", fontsize=size_axes)
    plt.ylabel(
        "Number of mRNAs detected (log scale)",
        fontweight="bold",
        fontsize=size_axes)
    if threshold is not None:
        plt.legend(prop={'size': size_legend})
    plt.tight_layout()
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


def plot_elbow_colocalized(spots_1, spots_2, voxel_size, threshold_max=None,
                           title=None, framesize=(5, 5), size_title=20,
                           size_axes=15, size_legend=15, path_output=None,
                           ext="png", show=True):
    """Plot the elbow curve that allows an automated colocalized spot
    detection.

    Parameters
    ----------
    spots_1 : np.ndarray
        Coordinates of the spots with shape (nb_spots, 3) or (nb_spots, 2).
    spots_2 : np.ndarray
        Coordinates of the spots with shape (nb_spots, 3) or (nb_spots, 2).
    voxel_size : int or float or tuple or list
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    threshold_max : int or float, optional
        Maximum threshold value to consider.
    title : str, optional
        Title of the plot.
    framesize : tuple, default=(5, 5)
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    size_title : int, default=20
        Size of the title.
    size_axes : int, default=15
        Size of the axes label.
    size_legend : int, default=15
        Size of the legend.
    path_output : str, optional
        Path to save the image (without extension).
    ext : str or list, default='png'
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool
        Show the figure or not.

    """
    # check parameters
    stack.check_parameter(
        threshold_max=(int, float, type(None)),
        title=(str, list, type(None)),
        framesize=tuple,
        size_title=int,
        size_axes=int,
        size_legend=int,
        path_output=(str, type(None)),
        ext=(str, list),
        show=bool)

    # get thresholds and colocalized spots count to plot the elbow curve
    (thresholds, count_colocalized,
     threshold) = multistack.get_elbow_value_colocalized(
        spots_1=spots_1,
        spots_2=spots_2,
        voxel_size=voxel_size)

    # plot
    plt.figure(figsize=framesize)
    plt.hlines(
        len(spots_1),
        thresholds[0],
        thresholds[-1],
        colors="forestgreen",
        linestyles="--",
        label="Spots 1")
    plt.hlines(
        len(spots_2),
        thresholds[0],
        thresholds[-1],
        colors="steelblue",
        linestyles="--",
        label="Spots 2")
    plt.plot(
        thresholds,
        count_colocalized,
        color="firebrick",
        label="Colocalized spots")
    if threshold is not None:
        i_threshold = np.argmax(thresholds == threshold)
        plt.scatter(
            threshold,
            count_colocalized[i_threshold],
            marker="D",
            c="#d7191c",
            s=60,
            label="Selected threshold")

    #  define xlim
    if threshold_max is not None:
        plt.xlim((0, threshold_max))

    # axes
    if title is not None:
        plt.title(title, fontweight="bold", fontsize=size_title)
    plt.xlabel("Thresholds", fontweight="bold", fontsize=size_axes)
    plt.ylabel(
        "Number of mRNAs detected",
        fontweight="bold",
        fontsize=size_axes)
    if threshold is not None:
        plt.legend(prop={'size': size_legend})
    plt.tight_layout()
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.detection.utils module.
"""

# TODO test apifish.detection.convert_spot_coordinates
# TODO add test for apifish.detection.get_object_radius_pixel
# TODO add test for apifish.detection.get_object_radius_nm
# TODO add test for apifish.detection.build_reference_spot
# TODO add test for apifish.detection.compute_snr_spots
# TODO add test for apifish.detection.get_breaking_point



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.segmentation.nuc_segmentation module.
"""


# TODO add test for apifish.segmentation.remove_segmented_nuc
# TODO add test for apifish.segmentation.unet_3_classes_nuc
# TODO add test for apifish.segmentation.apply_unet_3_classes
# TODO add test for apifish.segmentation.from_3_classes_to_instances



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Segmentation models.
"""

import os
import warnings

from zipfile import ZipFile

import tensorflow as tf

import apifish.stack as stack

from .utils_models import EncoderDecoder
from .utils_models import SameConv

from tensorflow.python.keras.layers import Input
from tensorflow.python.keras.layers import Softmax
from tensorflow.python.keras.layers import Concatenate
from tensorflow.python.keras.engine.training import Model


# ### Pre-trained models ###

def load_pretrained_model(channel, model_name):
    """Build and compile a model, then load its pretrained weights.

    Parameters
    ----------
    channel : str
        Input channel for the model ('nuc', 'cell' or 'double').
    model_name : str
        Name of the model used ('3-classes' or 'double_distance_edge).

    Returns
    -------
    model : tensorflow.keras.model object
        Pretrained Unet model.

    """
    # TODO fix warning partial restoration with distance model

    # check parameters
    stack.check_parameter(
        channel=str,
        model_name=str)

    # unet 3-classes for nucleus segmentation
    if model_name == "3_classes" and channel == "nuc":
        model = build_compile_3_classes_model()

    # unet distance map to edge for both nucleus and cell segmentation
    elif model_name == "distance_edge" and channel == "double":
        model = build_compile_double_distance_model()

    else:
        raise ValueError("Model name and channel to segment are not "
                         "consistent: {0} - {1}.".format(channel, model_name))

    # load weights
    path_pretrained_directory = check_pretrained_weights(channel, model_name)
    path_checkpoint = os.path.join(
        path_pretrained_directory, "checkpoint")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model.load_weights(path_checkpoint)

    return model


def check_pretrained_weights(channel, model_name):
    """Check pretrained weights exist and download them if necessary.

    Parameters
    ----------
    channel : str
        Input channel for the model ('nuc', 'cell' or 'double').
    model_name : str
        Name of the model used ('3-classes' or 'distance_edge').

    Returns
    -------
    directory_checkpoint : str
        Path of the checkpoint directory.

    """
    # check parameters
    stack.check_parameter(
        channel=str,
        model_name=str)

    # get path checkpoint
    path_weights_directory = _get_weights_directory()
    pretrained_directory = "_".join([channel, model_name])
    path_directory = os.path.join(path_weights_directory, pretrained_directory)
    download = False

    # get url and hash
    if channel == "nuc" and model_name == "3_classes":
        url_zip_file = "https://github.com/fish-quant/big-fish-examples/releases/download/weights_1/nuc_3_classes.zip"
        hash_checkpoint = "02988027faf3f16b4088ee83c2ade14098e8ffb325c23a576cc639dae48aa936"
        hash_data = "284d6b8cadb3eddd691f1407bfdd1a7e6fa085c57a2feac07446e84aa3b1baf8"
        hash_index = "c924f0f2be179340dc5c75e21833fd831d5a4efdfb48382edbbcff748a162bc4"
        hash_log = "b6d08b9fbacd3430d89eef2cd5c3246eb8c93e7b970ba17a76c8a9d72f58c8e7"
    elif channel == "double" and model_name == "distance_edge":
        url_zip_file = "https://github.com/fish-quant/big-fish-examples/releases/download/weights_0/double_distance_edge.zip"
        hash_checkpoint = "02988027faf3f16b4088ee83c2ade14098e8ffb325c23a576cc639dae48aa936"
        hash_data = "528aecbc6418df8d0d73fb23b3518ef779ae5135e15ca3e2e67518726f998d4d"
        hash_index = "cbebdd86868e46507733fccca589f940a438f4fde6f5d935e61f4112047db63e"
        hash_log = "63bc7e629d464bbf6cd3afd466f4d89a3be10edb2633de7ab825384a61326a09"
    else:
        raise ValueError("Model name and channel to segment are not "
                         "consistent: {0} - {1}.".format(channel, model_name))

    # case where pretrained directory exists
    if os.path.isdir(path_directory):

        # paths
        path_checkpoint = os.path.join(
            path_directory, "checkpoint")
        path_data = os.path.join(
            path_directory, "checkpoint.data-00000-of-00001")
        path_index = os.path.join(
            path_directory, "checkpoint.index")
        path_log = os.path.join(
            path_directory, "log")

        # checkpoint available and not corrupted
        if os.path.exists(path_checkpoint):
            try:
                stack.check_hash(path_checkpoint, hash_checkpoint)
            except IOError:
                print("{0} seems corrupted.".format(path_checkpoint))
                download = True
        else:
            download = True

        # data available and not corrupted
        if os.path.exists(path_data):
            try:
                stack.check_hash(path_data, hash_data)
            except IOError:
                print("{0} seems corrupted.".format(path_data))
                download = True
        else:
            download = True

        # index available and not corrupted
        if os.path.exists(path_index):
            try:
                stack.check_hash(path_index, hash_index)
            except IOError:
                print("{0} seems corrupted.".format(path_index))
                download = True
        else:
            download = True

        # log available and not corrupted
        if os.path.exists(path_log):
            try:
                stack.check_hash(path_log, hash_log)
            except IOError:
                print("{0} seems corrupted.".format(path_log))
                download = True
        else:
            download = True

    # case where pretrained directory does not exist
    else:
        download = True

    # download checkpoint files
    if download:
        print("downloading model weights...")

        # download zipfile
        stack.load_and_save_url(
            remote_url=url_zip_file,
            directory=path_weights_directory)

        # unzip
        path_zipfile = os.path.join(
            path_weights_directory, "{0}.zip".format(pretrained_directory))
        with ZipFile(path_zipfile, 'r') as z:
            z.extract(
                member="{0}/checkpoint".format(pretrained_directory),
                path=path_weights_directory)
            z.extract(
                member="{0}/checkpoint.data-00000-of-00001".format(
                    pretrained_directory),
                path=path_weights_directory)
            z.extract(
                member="{0}/checkpoint.index".format(pretrained_directory),
                path=path_weights_directory)
            z.extract(
                member="{0}/log".format(pretrained_directory),
                path=path_weights_directory)

        # check files consistency
        path = os.path.join(
            path_weights_directory,
            "{0}/checkpoint".format(pretrained_directory))
        stack.check_hash(path, hash_checkpoint)
        path = os.path.join(
            path_weights_directory,
            "{0}/checkpoint.data-00000-of-00001".format(pretrained_directory))
        stack.check_hash(path, hash_data)
        path = os.path.join(
            path_weights_directory,
            "{0}/checkpoint.index".format(pretrained_directory))
        stack.check_hash(path, hash_index)
        path = os.path.join(
            path_weights_directory,
            "{0}/log".format(pretrained_directory))
        stack.check_hash(path, hash_log)

        # remove zipfile
        os.remove(path_zipfile)

    return path_directory


def _get_weights_directory():
    path = os.path.realpath(__file__)
    path = os.path.dirname(path)
    path = os.path.dirname(path)
    path = os.path.join(path, "weights")

    return path


# ### Models 3-classes ###

def build_compile_3_classes_model():
    """Build and compile a Unet model to predict 3 classes from nucleus or
    cell images: background, edge and foreground.

    Returns
    -------
    model_3_classes : tensorflow.keras.model object
        Compiled Unet model.

    """
    # define inputs
    inputs_image = Input(
        shape=(None, None, 1),
        dtype="float32",
        name="image")

    # define model
    outputs = _get_3_classes_model(inputs_image)
    model_3_classes = Model(
        inputs_image,
        outputs,
        name="3ClassesModel")

    # losses
    loss = tf.keras.losses.SparseCategoricalCrossentropy()

    # metrics
    accuracy = tf.metrics.SparseCategoricalAccuracy(name="accuracy")

    # compile model
    model_3_classes.compile(
        optimizer='adam',
        loss=loss,
        metrics=accuracy)

    return model_3_classes


def _get_3_classes_model(inputs):
    """Build Unet architecture that return a 3-classes prediction.

    Parameters
    ----------
    inputs : tensorflow.keras.Input object
        Input layer with shape (B, H, W, 1) or (H, W, 1).

    Returns
    -------
    output : tensorflow.keras.layers object
        Output layer (softmax) with shape (B, H, W, 3) or (H, W, 3).

    """
    # compute feature map
    features_core = EncoderDecoder(
        name="encoder_decoder")(inputs)  # (B, H, W, 32)

    # compute 3-classes output
    features_3_classes = SameConv(
        filters=3,
        kernel_size=(1, 1),
        activation="linear",
        name="final_conv")(features_core)  # (B, H, W, 3)
    output = Softmax(
        axis=-1,
        name="label_3")(features_3_classes)  # (B, H, W, 3)

    return output


# ### Models distance map ###

def build_compile_distance_model():
    """Build and compile a Unet model to predict foreground and a distance map
    from nucleus or cell images.

    Returns
    -------
    model_distance : Tensorflow model
        Compiled Unet model.

    """
    # define inputs
    inputs_image = Input(
        shape=(None, None, 1),
        dtype="float32",
        name="image")

    # define model
    output_surface, output_distance = _get_distance_model(inputs_image)
    outputs = [output_surface, output_distance]
    model_distance = Model(
        inputs_image,
        outputs,
        name="DistanceModel")

    # losses
    loss_surface = tf.keras.losses.BinaryCrossentropy()
    loss_distance = tf.keras.losses.MeanAbsoluteError()
    losses = [[loss_surface], [loss_distance]]
    losses_weight = [[1.0], [1.0]]

    # metrics
    metric_surface = tf.metrics.BinaryAccuracy(name="accuracy")
    metric_distance = tf.metrics.MeanAbsoluteError(name="mae")
    metrics = [[metric_surface], [metric_distance]]

    # compile model
    model_distance.compile(
        optimizer='adam',
        loss=losses,
        loss_weights=losses_weight,
        metrics=metrics)

    return model_distance


def _get_distance_model(inputs):
    """Build Unet architecture that return a distance map.

    Parameters
    ----------
    inputs : tensorflow.keras.Input object
        Input layer with shape (B, H, W, 1) or (H, W, 1).

    Returns
    -------
    output_surface : tensorflow.keras.layers object
        Output layer for the foreground/background prediction with shape
        (B, H, W, 1) or (H, W, 1).
    output_distance : tensorflow.keras.layers object
        Output layer for the distance map with shape (B, H, W, 1) or (H, W, 1).

    """
    # compute feature map
    features_core = EncoderDecoder(
        name="encoder_decoder")(inputs)  # (B, H, W, 32)

    # compute surface output
    output_surface = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="sigmoid",
        name="label_2")(features_core)  # (B, H, W, 1)

    # compute distance output
    output_distance = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="relu",
        name="label_distance")(features_core)  # (B, H, W, 1)

    return output_surface, output_distance


def build_compile_double_distance_model():
    """Build and compile a Unet model to predict foreground and a distance map
    from nucleus and cell images.

    This model version takes two images as input (for nucleus and cell).

    Returns
    -------
    model_distance : Tensorflow model
        Compiled Unet model.

    """
    # define inputs
    inputs_nuc = Input(
        shape=(None, None, 1),
        dtype="float32",
        name="nuc")
    inputs_cell = Input(
        shape=(None, None, 1),
        dtype="float32",
        name="cell")
    inputs = [inputs_nuc, inputs_cell]

    # define model
    (output_distance_nuc, output_surface_cell,
     output_distance_cell) = _get_double_distance_model(inputs)
    outputs = [output_distance_nuc, output_surface_cell, output_distance_cell]
    model_distance = Model(
        inputs,
        outputs,
        name="DoubleDistanceModel")

    # losses
    loss_distance_nuc = tf.keras.losses.MeanAbsoluteError()
    loss_surface_cell = tf.keras.losses.BinaryCrossentropy()
    loss_distance_cell = tf.keras.losses.MeanAbsoluteError()
    losses = [[loss_distance_nuc],
              [loss_surface_cell], [loss_distance_cell]]
    losses_weight = [[1.0], [1.0], [1.0]]

    # metrics
    metric_distance_nuc = tf.metrics.MeanAbsoluteError(name="mae")
    metric_surface_cell = tf.metrics.BinaryAccuracy(name="accuracy")
    metric_distance_cell = tf.metrics.MeanAbsoluteError(name="mae")
    metrics = [[metric_distance_nuc],
               [metric_surface_cell], [metric_distance_cell]]

    # compile model
    model_distance.compile(
        optimizer='adam',
        loss=losses,
        loss_weights=losses_weight,
        metrics=metrics)

    return model_distance


def _get_double_distance_model(inputs):
    """Build Unet architecture that return nucleus and cell distance maps, and
    a cell surface prediction.

    Parameters
    ----------
    inputs : List[tensorflow.keras.Input object]
        List of two input layer with shape (B, H, W, 1) or (H, W, 1).

    Returns
    -------
    output_distance_nuc : tensorflow.keras.layers object
        Output layer for the nucleus distance map with shape (B, H, W, 1) or
        (H, W, 1).
    output_surface_cell : tensorflow.keras.layers object
        Output layer for the cell foreground/background prediction with shape
        (B, H, W, 1) or (H, W, 1).
    output_distance_cell : tensorflow.keras.layers object
        Output layer for the cell distance map with shape (B, H, W, 1) or
        (H, W, 1).

    """
    # compute feature map
    inputs_nuc, inputs_cell = inputs
    inputs = Concatenate(
        axis=-1)([inputs_nuc, inputs_cell])  # (B, H, W, 2)
    features_core = EncoderDecoder(
        name="encoder_decoder")(inputs)  # (B, H, W, 32)

    # compute distance output nucleus
    output_distance_nuc = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="relu",
        name="label_distance_nuc")(features_core)  # (B, H, W, 1)

    # compute surface output cell
    output_surface_cell = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="sigmoid",
        name="label_2_cell")(features_core)  # (B, H, W, 1)

    # compute distance output cell
    output_distance_cell = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="relu",
        name="label_distance_cell")(features_core)  # (B, H, W, 1)

    return output_distance_nuc, output_surface_cell, output_distance_cell



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.stack.augmentation module.
"""

import pytest

import numpy as np
import apifish.stack as stack

from apifish.stack.augmentation import _identity
from apifish.stack.augmentation import _flip_h
from apifish.stack.augmentation import _flip_v
from apifish.stack.augmentation import _transpose
from apifish.stack.augmentation import _transpose_inverse
from apifish.stack.augmentation import _rotation_90
from apifish.stack.augmentation import _rotation_180
from apifish.stack.augmentation import _rotation_270

from numpy.testing import assert_array_equal


# toy image
x = np.array([[1, 0, 0, 0, 0],
              [0, 1, 0, 0, 0],
              [0, 1, 0, 0, 0],
              [0, 1, 1, 1, 0],
              [0, 0, 0, 0, 0]], dtype=np.uint8)


def test_identity():
    y = _identity(x)
    assert_array_equal(y, x)


def test_flip_h():
    # one channel
    y = _flip_h(x)
    expected_y = np.array([[0, 0, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [1, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)

    # multichannel
    xx = x[..., np.newaxis]
    yy = _flip_h(xx)
    expected_yy = expected_y[..., np.newaxis]
    assert_array_equal(yy, expected_yy)


def test_flip_v():
    # one channel
    y = _flip_v(x)
    expected_y = np.array([[0, 0, 0, 0, 1],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)

    # multichannel
    xx = x[..., np.newaxis]
    yy = _flip_v(xx)
    expected_yy = expected_y[..., np.newaxis]
    assert_array_equal(yy, expected_yy)


def test_transpose():
    # one channel
    y = _transpose(x)
    expected_y = np.array([[1, 0, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)

    # multichannel
    xx = x[..., np.newaxis]
    yy = _transpose(xx)
    expected_yy = expected_y[..., np.newaxis]
    assert_array_equal(yy, expected_yy)


def test_transpose_inverse():
    y = _transpose_inverse(x)
    expected_y = np.array([[0, 0, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 0, 1]], dtype=np.uint8)
    assert_array_equal(y, expected_y)


def test_rotation_90():
    y = _rotation_90(x)
    expected_y = np.array([[0, 0, 0, 0, 1],
                           [0, 1, 1, 1, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)


def test_rotation_180():
    y = _rotation_180(x)
    expected_y = np.array([[0, 0, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 0, 1]], dtype=np.uint8)
    assert_array_equal(y, expected_y)


def test_rotation_270():
    y = _rotation_270(x)
    expected_y = np.array([[0, 0, 0, 0, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 1, 1, 1, 0],
                           [1, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32,np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64,
    bool])
def test_augment_2d_dtype(dtype):
    x = np.array([[1, 0, 0, 0, 0],
                  [0, 1, 0, 0, 0],
                  [0, 1, 0, 0, 0],
                  [0, 1, 1, 1, 0],
                  [0, 0, 0, 0, 0]], dtype=dtype)
    y = stack.augment_2d(x)
    assert y.dtype == dtype


def test_augment_2d_function():
    operations = [_identity,
                  _flip_h, _flip_v,
                  _transpose, _transpose_inverse,
                  _rotation_90, _rotation_180, _rotation_270]
    bytecodes = []
    for f in operations:
        bytecode = f.__code__.co_code
        bytecodes.append(bytecode)
    f = stack.augment_2d_function()
    assert f.__code__.co_code in bytecodes
    f = stack.augment_2d_function(identity=True)
    assert f.__code__.co_code == _identity.__code__.co_code


def test_augment_8_times():
    # one channel
    expected_y_identity = x.copy()
    expected_y_flip_h = np.array([[0, 0, 0, 0, 0],
                                  [0, 1, 1, 1, 0],
                                  [0, 1, 0, 0, 0],
                                  [0, 1, 0, 0, 0],
                                  [1, 0, 0, 0, 0]], dtype=np.uint8)
    expected_y_flip_v = np.array([[0, 0, 0, 0, 1],
                                  [0, 0, 0, 1, 0],
                                  [0, 0, 0, 1, 0],
                                  [0, 1, 1, 1, 0],
                                  [0, 0, 0, 0, 0]], dtype=np.uint8)
    expected_y_transpose = np.array([[1, 0, 0, 0, 0],
                                     [0, 1, 1, 1, 0],
                                     [0, 0, 0, 1, 0],
                                     [0, 0, 0, 1, 0],
                                     [0, 0, 0, 0, 0]], dtype=np.uint8)
    expected_y_transpose_inverse = np.array([[0, 0, 0, 0, 0],
                                             [0, 1, 0, 0, 0],
                                             [0, 1, 0, 0, 0],
                                             [0, 1, 1, 1, 0],
                                             [0, 0, 0, 0, 1]], dtype=np.uint8)
    expected_y_rotation_90 = np.array([[0, 0, 0, 0, 1],
                                       [0, 1, 1, 1, 0],
                                       [0, 1, 0, 0, 0],
                                       [0, 1, 0, 0, 0],
                                       [0, 0, 0, 0, 0]], dtype=np.uint8)
    expected_y_rotation_180 = np.array([[0, 0, 0, 0, 0],
                                        [0, 1, 1, 1, 0],
                                        [0, 0, 0, 1, 0],
                                        [0, 0, 0, 1, 0],
                                        [0, 0, 0, 0, 1]], dtype=np.uint8)
    expected_y_rotation_270 = np.array([[0, 0, 0, 0, 0],
                                        [0, 0, 0, 1, 0],
                                        [0, 0, 0, 1, 0],
                                        [0, 1, 1, 1, 0],
                                        [1, 0, 0, 0, 0]], dtype=np.uint8)
    expected_y = [expected_y_identity,
                  expected_y_flip_h, expected_y_flip_v,
                  expected_y_transpose, expected_y_transpose_inverse,
                  expected_y_rotation_90, expected_y_rotation_180,
                  expected_y_rotation_270]
    augmented_arrays = stack.augment_8_times(x)
    assert isinstance(augmented_arrays, list)
    assert len(augmented_arrays) == len(expected_y)
    for a, b in zip(augmented_arrays, expected_y):
        assert_array_equal(a, b)

    # multichannel
    xx = x[..., np.newaxis]
    expected_yy = [y[..., np.newaxis] for y in expected_y]
    augmented_arrays = stack.augment_8_times(xx)
    assert isinstance(augmented_arrays, list)
    assert len(augmented_arrays) == len(expected_yy)
    for a, b in zip(augmented_arrays, expected_yy):
        assert_array_equal(a, b)


def test_augment_8_times_reversed():
    y = stack.augment_8_times(x)
    y_reversed = stack.augment_8_times_reversed(y)
    assert isinstance(y_reversed, list)
    assert len(y_reversed) == 8
    for a in y_reversed:
        assert_array_equal(a, x.copy())



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.segmentation.postprocess module.
"""


# TODO add test for apifish.segmentation.label_instances
# TODO add test for apifish.segmentation.merge_labels
# TODO add test for apifish.segmentation.clean_segmentation
# TODO add test for apifish.segmentation.remove_disjoint
# TODO add test for apifish.segmentation.center_mask_coord
# TODO add test for apifish.segmentation.from_boundaries_to_surface
# TODO add test for apifish.segmentation.from_surface_to_boundaries
# TODO add test for apifish.segmentation.from_binary_to_coord
# TODO add test for apifish.segmentation.complete_coord_boundaries
# TODO add test for apifish.segmentation.from_coord_to_frame
# TODO add test for apifish.segmentation.from_coord_to_surface



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions to detect spots in 2-d and 3-d.
"""

import warnings

import scipy.ndimage as ndi
import numpy as np

import apifish.stack as stack

from .utils import get_object_radius_pixel
from .utils import get_breaking_point

from skimage.measure import regionprops
from skimage.measure import label


# ### Main function ###

def detect_spots(images, threshold=None, remove_duplicate=True,
                 return_threshold=False, voxel_size=None, spot_radius=None,
                 log_kernel_size=None, minimum_distance=None):
    """Apply LoG filter followed by a Local Maximum algorithm to detect spots
    in a 2-d or 3-d image.

    #. We smooth the image with a LoG filter.
    #. We apply a multidimensional maximum filter.
    #. A pixel which has the same value in the original and filtered images
       is a local maximum.
    #. We remove local peaks under a threshold.
    #. We keep only one pixel coordinate per detected spot.

    Parameters
    ----------
    images : List[np.ndarray] or np.ndarray
        Image (or list of images) with shape (z, y, x) or (y, x). If several
        images are provided, the same threshold is applied.
    threshold : int, float or None
        A threshold to discriminate relevant spots from noisy blobs. If None,
        optimal threshold is selected automatically. If several images are
        provided, one optimal threshold is selected for all the images.
    remove_duplicate : bool
        Remove potential duplicate coordinates for the same spots. Slow the
        running.
    return_threshold : bool
        Return the threshold used to detect spots.
    voxel_size : int, float, Tuple(int, float), List(int, float) or None
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions. Not used if 'log_kernel_size' and 'minimum_distance' are
        provided.
    spot_radius : int, float, Tuple(int, float), List(int, float) or None
        Radius of the spot, in nanometer. One value per spatial dimension (zyx
        or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions. Not used if 'log_kernel_size' and 'minimum_distance'
        are provided.
    log_kernel_size : int, float, Tuple(int, float), List(int, float) or None
        Size of the LoG kernel. It equals the standard deviation (in pixels)
        used for the gaussian kernel (one for each dimension). One value per
        spatial dimension (zyx or yx dimensions). If it's a scalar, the same
        standard deviation is applied to every dimensions. If None, we estimate
        it with the voxel size and spot radius.
    minimum_distance : int, float, Tuple(int, float), List(int, float) or None
        Minimum distance (in pixels) between two spots we want to be able to
        detect separately. One value per spatial dimension (zyx or yx
        dimensions). If it's a scalar, the same distance is applied to every
        dimensions. If None, we estimate it with the voxel size and spot
        radius.

    Returns
    -------
    spots : List[np.ndarray] or np.ndarray, np.int64
        Coordinates (or list of coordinates) of the spots with shape
        (nb_spots, 3) or (nb_spots, 2), for 3-d or 2-d images respectively.
    threshold : int or float
        Threshold used to discriminate spots from noisy blobs.

    """
    # check parameters
    stack.check_parameter(
        threshold=(int, float, type(None)),
        remove_duplicate=bool,
        return_threshold=bool,
        voxel_size=(int, float, tuple, list, type(None)),
        spot_radius=(int, float, tuple, list, type(None)),
        log_kernel_size=(int, float, tuple, list, type(None)),
        minimum_distance=(int, float, tuple, list, type(None)))

    # if one image is provided we enlist it
    if not isinstance(images, list):
        stack.check_array(
            images,
            ndim=[2, 3],
            dtype=[np.uint8, np.uint16, np.float32, np.float64])
        ndim = images.ndim
        images = [images]
        is_list = False
    else:
        ndim = None
        for i, image in enumerate(images):
            stack.check_array(
                image,
                ndim=[2, 3],
                dtype=[np.uint8, np.uint16, np.float32, np.float64])
            if i == 0:
                ndim = image.ndim
            else:
                if ndim != image.ndim:
                    raise ValueError("Provided images should have the same "
                                     "number of dimensions.")
        is_list = True

    # check consistency between parameters - detection with voxel size and
    # spot radius
    if (voxel_size is not None and spot_radius is not None
            and log_kernel_size is None and minimum_distance is None):
        if isinstance(voxel_size, (tuple, list)):
            if len(voxel_size) != ndim:
                raise ValueError("'voxel_size' must be a scalar or a sequence "
                                 "with {0} elements.".format(ndim))
        else:
            voxel_size = (voxel_size,) * ndim
        if isinstance(spot_radius, (tuple, list)):
            if len(spot_radius) != ndim:
                raise ValueError("'spot_radius' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            spot_radius = (spot_radius,) * ndim
        log_kernel_size = get_object_radius_pixel(
            voxel_size_nm=voxel_size,
            object_radius_nm=spot_radius,
            ndim=ndim)
        minimum_distance = get_object_radius_pixel(
            voxel_size_nm=voxel_size,
            object_radius_nm=spot_radius,
            ndim=ndim)

    # check consistency between parameters - detection with kernel size and
    # minimal distance
    elif (voxel_size is None and spot_radius is None
          and log_kernel_size is not None and minimum_distance is not None):
        if isinstance(log_kernel_size, (tuple, list)):
            if len(log_kernel_size) != ndim:
                raise ValueError("'log_kernel_size' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            log_kernel_size = (log_kernel_size,) * ndim
        if isinstance(minimum_distance, (tuple, list)):
            if len(minimum_distance) != ndim:
                raise ValueError("'minimum_distance' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            minimum_distance = (minimum_distance,) * ndim

    # check consistency between parameters - detection in priority with kernel
    # size and minimal distance
    elif (voxel_size is not None and spot_radius is not None
          and log_kernel_size is not None and minimum_distance is not None):
        if isinstance(log_kernel_size, (tuple, list)):
            if len(log_kernel_size) != ndim:
                raise ValueError("'log_kernel_size' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            log_kernel_size = (log_kernel_size,) * ndim
        if isinstance(minimum_distance, (tuple, list)):
            if len(minimum_distance) != ndim:
                raise ValueError("'minimum_distance' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            minimum_distance = (minimum_distance,) * ndim

    # missing parameters
    else:
        raise ValueError("One of the two pairs of parameters ('voxel_size', "
                         "'spot_radius') or ('log_kernel_size', "
                         "'minimum_distance') should be provided.")

    # detect spots
    if return_threshold:
        spots, threshold = _detect_spots_from_images(
            images,
            threshold=threshold,
            remove_duplicate=remove_duplicate,
            return_threshold=return_threshold,
            log_kernel_size=log_kernel_size,
            min_distance=minimum_distance)
    else:
        spots = _detect_spots_from_images(
            images,
            threshold=threshold,
            remove_duplicate=remove_duplicate,
            return_threshold=return_threshold,
            log_kernel_size=log_kernel_size,
            min_distance=minimum_distance)

    # format results
    if not is_list:
        spots = spots[0]

    # return threshold or not
    if return_threshold:
        return spots, threshold
    else:
        return spots


def _detect_spots_from_images(images, threshold=None, remove_duplicate=True,
                              return_threshold=False, log_kernel_size=None,
                              min_distance=None):
    """Apply LoG filter followed by a Local Maximum algorithm to detect spots
    in a 2-d or 3-d image.

    #. We smooth the image with a LoG filter.
    #. We apply a multidimensional maximum filter.
    #. A pixel which has the same value in the original and filtered images
       is a local maximum.
    #. We remove local peaks under a threshold.
    #. We keep only one pixel coordinate per detected spot.

    Parameters
    ----------
    images : List[np.ndarray]
        List of images with shape (z, y, x) or (y, x). The same threshold is
        applied to every images.
    threshold : float or int
        A threshold to discriminate relevant spots from noisy blobs. If None,
        optimal threshold is selected automatically. If several images are
        provided, one optimal threshold is selected for all the images.
    remove_duplicate : bool
        Remove potential duplicate coordinates for the same spots. Slow the
        running.
    return_threshold : bool
        Return the threshold used to detect spots.
    log_kernel_size : int, float, Tuple(int, float), List(int, float) or None
        Size of the LoG kernel. It equals the standard deviation (in pixels)
        used for the gaussian kernel (one for each dimension). One value per
        spatial dimension (zyx or yx dimensions). If it's a scalar, the same
        standard deviation is applied to every dimensions. If None, we estimate
        it with the voxel size and spot radius.
    min_distance : int, float, Tuple(int, float), List(int, float) or None
        Minimum distance (in pixels) between two spots we want to be able to
        detect separately. One value per spatial dimension (zyx or yx
        dimensions). If it's a scalar, the same distance is applied to every
        dimensions. If None, we estimate it with the voxel size and spot
        radius.

    Returns
    -------
    all_spots : List[np.ndarray], np.int64
        List of spot coordinates with shape (nb_spots, 3) or (nb_spots, 2),
        for 3-d or 2-d images respectively.
    threshold : int or float
        Threshold used to discriminate spots from noisy blobs.

    """
    # initialization
    n = len(images)

    # apply LoG filter and find local maximum
    images_filtered = []
    pixel_values = []
    masks = []
    for image in images:
        # filter image
        image_filtered = stack.log_filter(image, log_kernel_size)
        images_filtered.append(image_filtered)

        # get pixels value
        pixel_values += list(image_filtered.ravel())

        # find local maximum
        mask_local_max = local_maximum_detection(image_filtered, min_distance)
        masks.append(mask_local_max)

    # get optimal threshold if necessary based on all the images
    if threshold is None:

        # get threshold values we want to test
        thresholds = _get_candidate_thresholds(pixel_values)

        # get spots count and its logarithm
        all_value_spots = []
        minimum_threshold = float(thresholds[0])
        for i in range(n):
            image_filtered = images_filtered[i]
            mask_local_max = masks[i]
            spots, mask_spots = spots_thresholding(
                image_filtered, mask_local_max,
                threshold=minimum_threshold,
                remove_duplicate=False)
            value_spots = image_filtered[mask_spots]
            all_value_spots.append(value_spots)
        all_value_spots = np.concatenate(all_value_spots)
        thresholds, count_spots = _get_spot_counts(thresholds, all_value_spots)

        # select threshold where the kink of the distribution is located
        if count_spots.size > 0:
            threshold, _, _ = get_breaking_point(thresholds, count_spots)

    # detect spots
    all_spots = []
    for i in range(n):

        # get images and masks
        image_filtered = images_filtered[i]
        mask_local_max = masks[i]

        # detection
        spots, _ = spots_thresholding(image_filtered, mask_local_max,
                                      threshold, remove_duplicate)
        all_spots.append(spots)

    # return threshold or not
    if return_threshold:
        return all_spots, threshold
    else:
        return all_spots


# ### LoG spot detection ###

def local_maximum_detection(image, min_distance):
    """Compute a mask to keep only local maximum, in 2-d and 3-d.

    #. We apply a multidimensional maximum filter.
    #. A pixel which has the same value in the original and filtered images
       is a local maximum.

    Several connected pixels can have the same value. In such a case, the
    local maximum is not unique.

    In order to make the detection robust, it should be applied to a
    filtered image (using :func:`apifish.stack.log_filter` for example).

    Parameters
    ----------
    image : np.ndarray
        Image to process with shape (z, y, x) or (y, x).
    min_distance : int, float, Tuple(int, float), List(int, float)
        Minimum distance (in pixels) between two spots we want to be able to
        detect separately. One value per spatial dimension (zyx or yx
        dimensions). If it's a scalar, the same distance is applied to every
        dimensions.

    Returns
    -------
    mask : np.ndarray, bool
        Mask with shape (z, y, x) or (y, x) indicating the local peaks.

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_parameter(min_distance=(int, float, tuple, list))

    # compute the kernel size (centered around our pixel because it is uneven)
    if isinstance(min_distance, (tuple, list)):
        if len(min_distance) != image.ndim:
            raise ValueError(
                "'min_distance' should be a scalar or a sequence with one "
                "value per dimension. Here the image has {0} dimensions and "
                "'min_distance' {1} elements.".format(image.ndim,
                                                      len(min_distance)))
    else:
        min_distance = (min_distance,) * image.ndim
    min_distance = np.ceil(min_distance).astype(image.dtype)
    kernel_size = 2 * min_distance + 1

    # apply maximum filter to the original image
    image_filtered = ndi.maximum_filter(image, size=kernel_size)

    # we keep the pixels with the same value before and after the filtering
    mask = image == image_filtered

    return mask


def spots_thresholding(image, mask_local_max, threshold,
                       remove_duplicate=True):
    """Filter detected spots and get coordinates of the remaining spots.

    In order to make the thresholding robust, it should be applied to a
    filtered image (using :func:`apifish.stack.log_filter` for example). If
    the local maximum is not unique (it can happen if connected pixels have
    the same value), a connected component algorithm is applied to keep only
    one coordinate per spot.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    mask_local_max : np.ndarray, bool
        Mask with shape (z, y, x) or (y, x) indicating the local peaks.
    threshold : float, int or None
        A threshold to discriminate relevant spots from noisy blobs. If None,
        detection is aborted with a warning.
    remove_duplicate : bool
        Remove potential duplicate coordinates for the same spots. Slow the
        running.

    Returns
    -------
    spots : np.ndarray, np.int64
        Coordinate of the local peaks with shape (nb_peaks, 3) or
        (nb_peaks, 2) for 3-d or 2-d images respectively.
    mask : np.ndarray, bool
        Mask with shape (z, y, x) or (y, x) indicating the spots.

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_array(
        mask_local_max,
        ndim=[2, 3],
        dtype=[bool])
    stack.check_parameter(
        threshold=(float, int, type(None)),
        remove_duplicate=bool)

    if threshold is None:
        mask = np.zeros_like(image, dtype=bool)
        spots = np.array([], dtype=np.int64).reshape((0, image.ndim))
        warnings.warn("No spots were detected (threshold is {0})."
                      .format(threshold),
                      UserWarning)
        return spots, mask

    # remove peak with a low intensity
    mask = (mask_local_max & (image > threshold))
    if mask.sum() == 0:
        spots = np.array([], dtype=np.int64).reshape((0, image.ndim))
        return spots, mask

    # make sure we detect only one coordinate per spot
    if remove_duplicate:
        # when several pixels are assigned to the same spot, keep the centroid
        cc = label(mask)
        local_max_regions = regionprops(cc)
        spots = []
        for local_max_region in local_max_regions:
            spot = np.array(local_max_region.centroid)
            spots.append(spot)
        spots = np.stack(spots).astype(np.int64)

        # built mask again
        mask = np.zeros_like(mask)
        mask[spots[:, 0], spots[:, 1]] = True

    else:
        # get peak coordinates
        spots = np.nonzero(mask)
        spots = np.column_stack(spots)

    # case where no spots were detected
    if spots.size == 0:
        warnings.warn("No spots were detected (threshold is {0})."
                      .format(threshold),
                      UserWarning)

    return spots, mask


# ### Threshold selection ###

def automated_threshold_setting(image, mask_local_max):
    """Automatically set the optimal threshold to detect spots.

    In order to make the thresholding robust, it should be applied to a
    filtered image (using :func:`apifish.stack.log_filter` for example). The
    optimal threshold is selected based on the spots distribution. The latter
    should have an elbow curve discriminating a fast decreasing stage from a
    more stable one (a plateau).

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    mask_local_max : np.ndarray, bool
        Mask with shape (z, y, x) or (y, x) indicating the local peaks.

    Returns
    -------
    optimal_threshold : int
        Optimal threshold to discriminate spots from noisy blobs.

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_array(mask_local_max, ndim=[2, 3], dtype=[bool])

    # get threshold values we want to test
    thresholds = _get_candidate_thresholds(image.ravel())

    # get spots count and its logarithm
    first_threshold = float(thresholds[0])
    spots, mask_spots = spots_thresholding(
        image, mask_local_max, first_threshold, remove_duplicate=False)
    value_spots = image[mask_spots]
    thresholds, count_spots = _get_spot_counts(thresholds, value_spots)

    # select threshold where the break of the distribution is located
    if count_spots.size > 0:
        optimal_threshold, _, _ = get_breaking_point(thresholds, count_spots)

    # case where no spots were detected
    else:
        optimal_threshold = None

    return optimal_threshold


def _get_candidate_thresholds(pixel_values):
    """Choose the candidate thresholds to test for the spot detection.

    Parameters
    ----------
    pixel_values : np.ndarray
        Pixel intensity values of the image.

    Returns
    -------
    thresholds : np.ndarray, np.float64
        Candidate threshold values.

    """
    # choose appropriate thresholds candidate
    start_range = 0
    end_range = int(np.percentile(pixel_values, 99.9999))
    if end_range < 100:
        thresholds = np.linspace(start_range, end_range, num=100)
    else:
        thresholds = [i for i in range(start_range, end_range + 1)]
    thresholds = np.array(thresholds)

    return thresholds


def _get_spot_counts(thresholds, value_spots):
    """Compute and format the spots count function for different thresholds.

    Parameters
    ----------
    thresholds : np.ndarray, np.float64
        Candidate threshold values.
    value_spots : np.ndarray
        Pixel intensity values of all spots.

    Returns
    -------
    thresholds : np.ndarray, np.float64
        Candidate threshold values.
    count_spots : np.ndarray, np.float64
        Spots count function (log scale).

    """
    # count spots for each threshold
    count_spots = np.log([np.count_nonzero(value_spots > t)
                          for t in thresholds])
    count_spots = stack.centered_moving_average(count_spots, n=5)

    # the tail of the curve unnecessarily flatten the slop
    count_spots = count_spots[count_spots > 2]
    thresholds = thresholds[:count_spots.size]

    return thresholds, count_spots


def get_elbow_values(images, voxel_size=None, spot_radius=None,
                     log_kernel_size=None, minimum_distance=None):
    """Get values to plot the elbow curve used to automatically set the
    threshold to detect spots.

    Parameters
    ----------
    images : List[np.ndarray] or np.ndarray
        Image (or list of images) with shape (z, y, x) or (y, x). If several
        images are provided, the same threshold is applied.
    voxel_size : int, float, Tuple(int, float), List(int, float) or None
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions. Not used if 'log_kernel_size' and 'minimum_distance' are
        provided.
    spot_radius : int, float, Tuple(int, float), List(int, float) or None
        Radius of the spot, in nanometer. One value per spatial dimension (zyx
        or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions. Not used if 'log_kernel_size' and 'minimum_distance'
        are provided.
    log_kernel_size : int, float, Tuple(int, float), List(int, float) or None
        Size of the LoG kernel. It equals the standard deviation (in pixels)
        used for the gaussian kernel (one for each dimension). One value per
        spatial dimension (zyx or yx dimensions). If it's a scalar, the same
        standard deviation is applied to every dimensions. If None, we estimate
        it with the voxel size and spot radius.
    minimum_distance : int, float, Tuple(int, float), List(int, float) or None
        Minimum distance (in pixels) between two spots we want to be able to
        detect separately. One value per spatial dimension (zyx or yx
        dimensions). If it's a scalar, the same distance is applied to every
        dimensions. If None, we estimate it with the voxel size and spot
        radius.

    Returns
    -------
    thresholds : np.ndarray, np.float64
        Candidate threshold values.
    count_spots : np.ndarray, np.float64
        Spots count (log scale).
    threshold : float or None
        Threshold automatically set.

    """
    # check parameters
    stack.check_parameter(
        voxel_size=(int, float, tuple, list, type(None)),
        spot_radius=(int, float, tuple, list, type(None)),
        log_kernel_size=(int, float, tuple, list, type(None)),
        minimum_distance=(int, float, tuple, list, type(None)))

    # if one image is provided we enlist it
    if not isinstance(images, list):
        stack.check_array(
            images,
            ndim=[2, 3],
            dtype=[np.uint8, np.uint16, np.float32, np.float64])
        ndim = images.ndim
        images = [images]
        n = 1
    else:
        ndim = None
        for i, image in enumerate(images):
            stack.check_array(
                image,
                ndim=[2, 3],
                dtype=[np.uint8, np.uint16, np.float32, np.float64])
            if i == 0:
                ndim = image.ndim
            else:
                if ndim != image.ndim:
                    raise ValueError("Provided images should have the same "
                                     "number of dimensions.")
        n = len(images)

    # check consistency between parameters - detection with voxel size and
    # spot radius
    if (voxel_size is not None and spot_radius is not None
            and log_kernel_size is None and minimum_distance is None):
        if isinstance(voxel_size, (tuple, list)):
            if len(voxel_size) != ndim:
                raise ValueError(
                    "'voxel_size' must be a scalar or a sequence "
                    "with {0} elements.".format(ndim))
        else:
            voxel_size = (voxel_size,) * ndim
        if isinstance(spot_radius, (tuple, list)):
            if len(spot_radius) != ndim:
                raise ValueError("'spot_radius' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            spot_radius = (spot_radius,) * ndim

        log_kernel_size = get_object_radius_pixel(
            voxel_size_nm=voxel_size,
            object_radius_nm=spot_radius,
            ndim=ndim)
        minimum_distance = get_object_radius_pixel(
            voxel_size_nm=voxel_size,
            object_radius_nm=spot_radius,
            ndim=ndim)

    # check consistency between parameters - detection with kernel size and
    # minimal distance
    elif (voxel_size is None and spot_radius is None
          and log_kernel_size is not None and minimum_distance is not None):
        if isinstance(log_kernel_size, (tuple, list)):
            if len(log_kernel_size) != ndim:
                raise ValueError("'log_kernel_size' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            log_kernel_size = (log_kernel_size,) * ndim
        if isinstance(minimum_distance, (tuple, list)):
            if len(minimum_distance) != ndim:
                raise ValueError(
                    "'minimum_distance' must be a scalar or a "
                    "sequence with {0} elements.".format(ndim))
        else:
            minimum_distance = (minimum_distance,) * ndim

    # check consistency between parameters - detection in priority with kernel
    # size and minimal distance
    elif (voxel_size is not None and spot_radius is not None
          and log_kernel_size is not None and minimum_distance is not None):
        if isinstance(log_kernel_size, (tuple, list)):
            if len(log_kernel_size) != ndim:
                raise ValueError("'log_kernel_size' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            log_kernel_size = (log_kernel_size,) * ndim
        if isinstance(minimum_distance, (tuple, list)):
            if len(minimum_distance) != ndim:
                raise ValueError(
                    "'minimum_distance' must be a scalar or a "
                    "sequence with {0} elements.".format(ndim))
        else:
            minimum_distance = (minimum_distance,) * ndim

    # missing parameters
    else:
        raise ValueError(
            "One of the two pairs of parameters ('voxel_size', "
            "'spot_radius') or ('log_kernel_size', "
            "'minimum_distance') should be provided.")

    # apply LoG filter and find local maximum
    images_filtered = []
    pixel_values = []
    masks = []
    for image in images:
        # filter image
        image_filtered = stack.log_filter(image, log_kernel_size)
        images_filtered.append(image_filtered)

        # get pixels value
        pixel_values += list(image_filtered.ravel())

        # find local maximum
        mask_local_max = local_maximum_detection(
            image_filtered, minimum_distance)
        masks.append(mask_local_max)

    # get threshold values we want to test
    thresholds = _get_candidate_thresholds(pixel_values)

    # get spots count and its logarithm
    all_value_spots = []
    minimum_threshold = float(thresholds[0])
    for i in range(n):
        image_filtered = images_filtered[i]
        mask_local_max = masks[i]
        spots, mask_spots = spots_thresholding(
            image_filtered, mask_local_max,
            threshold=minimum_threshold,
            remove_duplicate=False)
        value_spots = image_filtered[mask_spots]
        all_value_spots.append(value_spots)
    all_value_spots = np.concatenate(all_value_spots)
    thresholds, count_spots = _get_spot_counts(
        thresholds, all_value_spots)

    # select threshold where the kink of the distribution is located
    if count_spots.size > 0:
        threshold, _, _ = get_breaking_point(thresholds, count_spots)
    else:
        threshold = None

    return thresholds, count_spots, threshold
